from .lfp_viewpoints import LfpViewpoints
from .lfp_exporter import LfpExporter
from .lfp_color_eq import LfpColorEqualizer
from .lfp_cropper import LfpCropper
from .lfp_depth import LfpDepth
from .lfp_rearranger import LfpRearranger
from .lfp_contrast import LfpContrast
from .lfp_outliers import LfpOutliers
from .hex_corrector import HexCorrector
from .top_level import LfpExtractor



from tests.unit_test_custom import PlenoptiCamTesterCustom
from tests.unit_test_illum import PlenoptiCamTesterIllum
from tests.unit_test_calib import PlenoptiCamTesterCalib
from tests.unit_test_cli import PlenoptiCamTesterCli
from tests.unit_test_gui import PlenoptiCamTesterGui
from tests.unit_test_err import PlenoptiCamErrorTester
from tests.unit_test_plt import PlenopticamTesterPlt

test_classes = [PlenoptiCamTesterCustom, PlenoptiCamTesterIllum, PlenoptiCamTesterCalib,
                PlenoptiCamTesterCli, PlenoptiCamTesterGui, PlenoptiCamErrorTester, PlenopticamTesterPlt]

for test_class in test_classes:
    obj = test_class()
    obj.setUp()
    obj.test_all()
    del obj



def bresenham_circle(arr_dims: (int, int) = (3, 3), r: int = 1) -> list:
    """ draw circle based on Bresenham's method """

    mask = [[0] * arr_dims[1] for _ in range(arr_dims[0])]
    xc, yc = arr_dims[1]//2, arr_dims[0]//2

    d = 3 - 2*r
    x = 0
    y = r
    mask = draw_circle(mask, xc, yc, x, y)

    while y >= x:

        # for each pixel we will draw all eight pixels
        x += 1

        # check for decision parameter and correspondingly update d, x, y
        if d > 0:
            y -= 1
            d = d + 4 * (x - y) + 10
        else:
            d = d + 4 * x + 6
        mask = draw_circle(mask, xc, yc, x, y)

    return mask


def draw_circle(mask: list = None, xc: int = 0, yc: int = 0, x: int = 0, y: int = 0):

    mask[xc+x][yc+y] = 1
    mask[xc-x][yc+y] = 1
    mask[xc+x][yc-y] = 1
    mask[xc-x][yc-y] = 1
    mask[xc+y][yc+x] = 1
    mask[xc-y][yc+x] = 1
    mask[xc+y][yc-x] = 1
    mask[xc-y][yc-x] = 1

    return mask



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import operator

from plenopticam.lfp_calibrator.find_centroid import find_centroid
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc.status import PlenopticamStatus


class CentroidSorter(object):

    def __init__(self, centroids, cfg=None, sta=None, bbox=None):
        super(CentroidSorter, self).__init__()

        # input variables
        self._centroids = np.asarray(centroids)     # list of unsorted maxima
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()
        self._bbox = bbox if bbox is not None and len(bbox) == 2 else None

        # internal variables
        self._lens_x_max = None  # maximum number of "complete" micro image centers in a row
        self._lens_y_max = None  # maximum number of "complete" micro image centers in a column
        self._upper_l = None     # upper left centroid
        self._lower_r = None     # lower right centroid
        self._upper_r = None     # upper right centroid
        self._lower_l = None     # lower left centroid

        # output variables
        self._mic_list = []                     # list of micro image centers with indices assigned
        self._pattern = None                    # pattern typ of micro lens array
        self._pitch = None                      # average pitch in horizontal and vertical direction
        self._hex_odd = True                    # hexagonal shift of second row on the right (True) or left (False)

        # initialize parameters
        self._init_var()

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        try:
            # get maximum number of micro images in horizontal and vertical direction
            self._mla_dims()
        except IndexError:
            self.sta.status_msg(msg="Error in MLA dimension estimation", opt=self.cfg.params[self.cfg.opt_prnt])
            self.sta.error = True
            return False

        # sort MICs and assign 2-D indices to them
        self._assign_mic_idx()

        # attach results to config object if present
        if hasattr(self, 'cfg') and hasattr(self.cfg, 'calibs'):
            self.cfg.calibs[self.cfg.mic_list] = self._mic_list
            self.cfg.calibs[self.cfg.pat_type] = self._pattern
            self.cfg.calibs[self.cfg.ptc_mean] = self._pitch

        return True

    def _init_var(self):
        """ find most up-left and most bottom-right micro image centers """

        # iterate through all micro image centers and add up x and y coordinates
        sum_mic = sum([self._centroids[:, 0], self._centroids[:, 1]])

        # center coordinates of micro image being top and most left
        self._upper_l = self._centroids[np.argmin(sum_mic)]

        # center coordinates of micro image being bottom and most right
        self._lower_r = self._centroids[np.argmax(sum_mic)]

        # set bounding box of micro image center field
        self._bbox = (self._lower_r - self._upper_l).astype('int') if not self._bbox else self._bbox

        # estimate micro image pitch lengths and pattern type
        self._get_mla_pitch()

        # determine if hexagonal shift of second row on the right (True) or left (False)
        self._hex_odd = not self.estimate_odd(self._upper_l, 0, inv_dir=0)

        return True

    def _mla_dims(self, counterclockwise_opt: bool = True):
        """ search for complete rows and columns and count number of centroids in each direction """

        y_max_l, y_max_r, x_max_t, x_max_b = [0, 0, 0, 0]
        self._upper_r = self._lower_r
        self._lower_l = self._upper_l

        # walk along rectangle in a clock-wise manner to find valid corners with complete row/col (twice for upper_l)
        for _ in range(2):
            # top row (west to east)
            x_max_t, self._upper_r, self._upper_l = self._get_lens_count(self._upper_l, self._upper_r, axis=1, inv_dir=0, inwards=0)
            # most right column (north to south)
            y_max_r, self._lower_r, self._upper_r = self._get_lens_count(self._upper_r, self._lower_r, axis=0, inv_dir=0, inwards=1)
            # bottom row (east to west)
            x_max_b, self._lower_l, self._lower_r = self._get_lens_count(self._lower_r, self._lower_l, axis=1, inv_dir=1, inwards=1)
            # most left column (south to north)
            y_max_l, self._upper_l, self._lower_l = self._get_lens_count(self._lower_l, self._upper_l, axis=0, inv_dir=1, inwards=0)

        # counter-clockwise
        if counterclockwise_opt:
            # most left column (north to south)
            y_max_l, self._lower_l, self._upper_l = self._get_lens_count(self._upper_l, self._lower_l, axis=0, inv_dir=0, inwards=0)
            # bottom row (west to east)
            x_max_b, self._lower_r, self._lower_l = self._get_lens_count(self._lower_l, self._lower_r, axis=1, inv_dir=0, inwards=1)
            # most right column (south to north)
            y_max_r, self._upper_r, self._lower_r = self._get_lens_count(self._lower_r, self._upper_r, axis=0, inv_dir=1, inwards=1)
            # top row (east to west)
            x_max_t, self._upper_l, self._upper_r = self._get_lens_count(self._upper_r, self._upper_l, axis=1, inv_dir=1, inwards=0)

        # set safe number of micro lenses in each direction
        self._lens_y_max, self._lens_x_max = min(y_max_l, y_max_r), min(x_max_t, x_max_b)

        return self._lens_y_max, self._lens_x_max

    def _assign_mic_idx(self):

        # reorder MIC list by neighbour relationship
        last_neighbor = self._upper_l
        self._mic_list = []
        self._mic_list.append([last_neighbor[0], last_neighbor[1], 0, 0])
        j = 0
        odd = self._hex_odd
        row_len_odd = True

        # print status
        self.sta.status_msg('Sort micro image centers', self.cfg.params[self.cfg.opt_prnt])

        # jump to the next row
        for ly in range(self._lens_y_max):
            # find all centers in one row
            for lx in range(self._lens_x_max-1):    # -1 to account for first row center which is already found
                # get adjacent MIC
                found_center = find_centroid(self._centroids, last_neighbor, self._pitch, 1, 'rec', None)
                # retrieve single MIC
                if len(found_center) != 2:
                    if len(found_center) > 2:
                        # average of found centroids
                        found_center = np.mean(found_center.reshape(-1, 2), axis=0)
                    else:
                        # skip when looking for last MIC with unequal row length
                        if lx == self._lens_x_max-1 and (row_len_odd or self._pattern == 'rec'):
                            break
                        # create missing centroid
                        found_center = self._mic_list[j][:2]+np.array([0, self._pitch[1]])
                j += 1
                self._mic_list.append([found_center[0], found_center[1], ly, lx+1])
                last_neighbor = found_center
            # find most-left center of next row
            if ly < self._lens_y_max-1:
                # get adjacent MIC
                found_center = find_centroid(self._centroids, self._upper_l, self._pitch, 0, self._pattern, odd)
                # retrieve single MIC
                if len(found_center) != 2:
                    if len(found_center) > 2:
                        # average of found centroids
                        found_center = np.mean(found_center.reshape(-1, 2), axis=0)
                    else:
                        # create missing centroid (considering MLA packing)
                        if self._pattern == 'rec':
                            found_center = self._upper_l + np.array([self._pitch[0], 0])
                        elif self._pattern == 'hex':
                            if odd:
                                found_center = self._upper_l + np.array([self._pitch[0], +self._pitch[1]/2])
                            else:
                                found_center = self._upper_l + np.array([self._pitch[0], -self._pitch[1]/2])
                j += 1
                self._mic_list.append([found_center[0], found_center[1], ly+1, 0])
                last_neighbor = found_center
                self._upper_l = found_center
                odd = not odd

            # check interrupt status
            if self.sta.interrupt:
                return False

            # print progress status on console
            self.sta.progress((ly+1)/self._lens_y_max*100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def _estimate_mla_geometry(self, pitch: float = None):
        """ This function determines whether the geometric arrangement of micro lenses is rectangular or hexagonal.

        :param pitch: scalar of type int or float representing the spacing between micro image centers
        :return: pattern, e.g. 'hex' or 'rec'
        """

        # pick arbitrary micro image center in middle of centroid list
        ref_point = self._centroids[len(self._centroids)//2][:2]

        # obtain list of adjacent centroids
        euc_dst = np.sum((self._centroids[:, :2]-ref_point)**2, axis=1)**.5
        adj_cnd = (pitch*.5 < euc_dst) & (pitch*1.5 > euc_dst)
        adj_pts = self._centroids[:, :2][adj_cnd]

        # place adjacent points at origin
        vecs = np.abs(adj_pts-ref_point)

        # create reference vectors at 45 degrees
        refs = np.repeat(np.array([1, 1])[..., None], vecs.shape[0], axis=1).T

        # element-wise dot product for angles
        angles = np.arccos(np.einsum('ij,ij->i', refs, vecs)/np.sqrt(2*np.einsum('ij,ij->i', vecs, vecs)))*180/np.pi

        # numeric angle reduction to account for tolerances (0 => 45 degs; 1 => 30 degs; 2 => 15 degs; 3 => 0 degs)
        angles = np.rint(angles / 15)

        # hexagonal angles
        if (sum(angles == 1) >= 1 and sum(angles == 3) >= 1) or (sum(angles == 2) >= 1 and sum(angles == 0) >= 1):
            self._pattern = 'hex'
        # rectangular angles
        elif sum(angles == 0) >= 0 and sum(angles == 3) >= 1:
            self._pattern = 'rec'
        else:
            self.sta.status_msg("Geometric MLA arrangement not recognized", opt=self.cfg.params[self.cfg.opt_prnt])
            self.sta.error = True
            return False

        return self._pattern

    def _get_mla_pitch(self):
        """
        get horizontal spacing estimate (pitch in px)
        """

        # use aspect ratio for "many" micro images
        if len(self._centroids) > 30**2:

            # get aspect ratio of bounding box
            aspect_ratio = self._bbox[1] / self._bbox[0]

            # get micro lens array dimensions
            J = np.sqrt(len(self._centroids) * aspect_ratio**-1)
            H = np.sqrt(len(self._centroids) * aspect_ratio)

            # use aspect ratio for pitch estimate
            pitch_x = self._bbox[1] / H

        # for fewer micro images use average pitch analysis
        else:
            pitch_diff = np.diff(self._centroids[:, 1])
            pitch_diff = pitch_diff[pitch_diff > 0]
            pitch_x = np.mean(pitch_diff[(pitch_diff <= np.mean(pitch_diff) + np.std(pitch_diff)) &
                                         (pitch_diff >= np.mean(pitch_diff) - np.std(pitch_diff))])

        # estimate MLA packing geometry
        self._estimate_mla_geometry(pitch_x)

        # get vertical spacing estimate (pitch in px) under consideration of packing type
        pitch_y = pitch_x * np.sqrt(3)/2 if self._pattern == 'hex' else pitch_x

        # store pitch values
        self._pitch = [pitch_y, pitch_x]

        return True

    def _get_lens_count(self, start_mic, opposite_mic, axis=0, inv_dir=0, inwards=0):

        cur_mic = start_mic
        lens_max = 1    # start to count from 1 to take existing centroid into account
        odd = not (self._hex_odd and not inwards) if not inv_dir else not inwards
        start_odd = odd
        comp_a = operator.lt if inv_dir else operator.gt
        pm_a = operator.add if inv_dir else operator.sub

        # iterate through row or column (as long as possible)
        while True:
            # get adjacent MIC
            found_center = find_centroid(self._centroids, cur_mic, self._pitch, axis=axis,
                                         pattern=self._pattern, odd=odd, inv_dir=inv_dir)
            if len(found_center) != 2:
                # if several candidates are found
                if len(found_center) > 2:
                    # average found centroids
                    found_center = np.mean(found_center.reshape(-1, 2), axis=0)
                # if stop condition is fulfilled
                elif comp_a(cur_mic[axis], pm_a(opposite_mic[axis], self._pitch[axis]/2)):
                    break
                else:
                    # restart with new row / column (extend search window e to ensure new start_mic is found)
                    start_mic = find_centroid(self._centroids, start_mic, self._pitch, axis=not axis,
                                              pattern=self._pattern, odd=start_odd, e=3.5, inv_dir=inwards)
                    start_odd = not start_odd   # reset hex search direction (flip for next not odd)
                    found_center = start_mic
                    lens_max = 0    # assign zero as number is incremented directly hereafter
                    odd = start_odd
            lens_max += 1
            cur_mic = found_center
            odd = not odd

        return lens_max, cur_mic, start_mic

    def estimate_odd(self, mic, axis, inv_dir=0):
        # look if hex shift in next row/col is left/top of the MIC and yield True if so
        return find_centroid(self._centroids, mic, self._pitch, axis, 'hex', odd=0, inv_dir=inv_dir).size == 0

    @property
    def mic_list(self):
        return self._mic_list

    @property
    def pattern(self):
        return self._pattern

    @property
    def pitch(self):
        return self._pitch



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# external
from color_space_converter import rgb2gry

# local imports
from plenopticam.lfp_calibrator.pitch_estimator import PitchEstimator
from plenopticam.lfp_calibrator.centroid_extractor import CentroidExtractor
from plenopticam.lfp_calibrator.centroid_refiner import CentroidRefiner
from plenopticam.lfp_calibrator.centroid_sorter import CentroidSorter
from plenopticam.lfp_calibrator.centroid_drawer import CentroidDrawer
from plenopticam.lfp_calibrator.grid_fitter import GridFitter
from plenopticam.lfp_calibrator.centroid_fit_sort import CentroidFitSorter
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.lfp_aligner.cfa_processor import CfaProcessor
from plenopticam.cfg import constants as c


class LfpCalibrator(object):

    def __init__(self, wht_img, cfg=None, sta=None):

        # input variables
        self._wht_img = wht_img
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()

        # private
        self._M = None

    def main(self):

        if self._wht_img is None:
            self.sta.status_msg(msg='White image file not present', opt=self.cfg.params[self.cfg.opt_prnt])
            self.sta.error = True

        # convert Bayer to RGB representation
        if len(self._wht_img.shape) == 2 and 'bay' in self.cfg.lfpimg:
            # perform color filter array management and obtain rgb image
            cfa_obj = CfaProcessor(bay_img=self._wht_img, cfg=self.cfg, sta=self.sta)
            cfa_obj.bay2rgb()
            self._wht_img = cfa_obj.rgb_img
            del cfa_obj

        # ensure white image is monochromatic
        if len(self._wht_img.shape) == 3:
            self._wht_img = rgb2gry(self._wht_img)[..., 0] if self._wht_img.shape[-1] == 3 else self._wht_img

        # estimate micro image diameter
        obj = PitchEstimator(self._wht_img, self.cfg, self.sta)
        obj.main()
        self._M = obj.M if self._M is None else self._M
        del obj

        # compute all centroids of micro images
        obj = CentroidExtractor(self._wht_img, self.cfg, self.sta, self._M)
        obj.main()
        centroids = obj.centroids
        peak_img = obj.peak_img
        del obj

        # refine centroids with sub-pixel precision using provided method
        obj = CentroidRefiner(peak_img, centroids, self.cfg, self.sta, self._M)
        obj.main()
        centroids = obj.centroids_refined
        del obj

        # obtain consistent MLA dimensions with indices for each micro image
        if self.cfg.params[self.cfg.cal_meth] == c.CALI_METH[4]:
            # perform 4-point corner fit for dimension and index assignment
            obj = CentroidFitSorter(centroids, self.cfg, self.sta)
            mic_list = obj.corner_fit()
            pattern, pitch = obj.pattern, obj.pitch
            del obj
        else:
            # iteratively reorder MICs and assign indices
            obj = CentroidSorter(centroids, self.cfg, self.sta)
            obj.main()
            mic_list, pattern, pitch = obj.mic_list, obj.pattern, obj.pitch
            del obj

        # fit grid of MICs using least-squares method to obtain accurate MICs from line intersections
        if self.cfg.params[self.cfg.cal_meth] in c.CALI_METH[2:4] and not self.sta.interrupt:
            obj = GridFitter(coords_list=mic_list, cfg=self.cfg, sta=self.sta, arr_shape=self._wht_img.shape,
                             pat_type=pattern, penalty_enable=self.cfg.params[self.cfg.cal_meth] == c.CALI_METH[3])
            obj.main()
            mic_list = obj.grid_fit
            del obj

        # attach calibration results to config object
        for kw, data in zip([self.cfg.mic_list, self.cfg.pat_type, self.cfg.ptc_mean], [mic_list, pattern, pitch]):
            self.cfg.calibs[kw] = data

        # save calibration metadata
        self.sta.status_msg('Save calibration data', opt=self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, opt=self.cfg.params[self.cfg.opt_prnt])
        try:
            self.cfg.save_cal_data(mic_list=mic_list, pat_type=pattern, ptc_mean=pitch)
            self.sta.progress(100, opt=self.cfg.params[self.cfg.opt_prnt])
        except PermissionError:
            self.sta.status_msg('Could not save calibration data', opt=self.cfg.params[self.cfg.opt_prnt])

        # write image to hard drive (only if debug option is set)
        if self.cfg.params[self.cfg.opt_dbug]:
            draw_obj = CentroidDrawer(self._wht_img, mic_list, self.cfg, self.sta)
            draw_obj.write_centroids_img(fn='wht_img+mics_sorted.png')
            del draw_obj

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
import unittest

from plenopticam.misc import PlenopticamError
from plenopticam.misc import PlenopticamStatus
from plenopticam.cfg import PlenopticamConfig
from plenopticam.lfp_reader import LfpReader
from plenopticam.misc import rm_file, rmdir_p


class PlenoptiCamErrorTester(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(PlenoptiCamErrorTester, self).__init__(*args, **kwargs)

    def setUp(self):

        self.cfg = PlenopticamConfig()
        self.sta = PlenopticamStatus()

    def test_read_error(self):

        # folder and path handling
        fp = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'examples', 'data')
        os.makedirs(fp) if not os.path.exists(fp) else None

        # create dummy file with wrong file format
        self.cfg.params[self.cfg.lfp_path] = os.path.join(fp, 'test_dummy.lfp')
        with open(self.cfg.params[self.cfg.lfp_path], 'a'):
            os.utime(self.cfg.params[self.cfg.lfp_path], None)

        with self.assertRaises(PlenopticamError) as exc:
            reader = LfpReader(cfg=self.cfg, sta=self.sta)
            reader.main()

        self.assertEqual("'dict' object has no attribute 'startswith'", str(exc.exception))

        # remove dummy data after test
        rm_file(self.cfg.params[self.cfg.lfp_path])
        rmdir_p(self.cfg.exp_path)

    def test_all(self):

        self.test_read_error()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.lfp_calibrator.non_max_supp import NonMaxSuppression
from plenopticam.lfp_calibrator.centroid_drawer import CentroidDrawer
from plenopticam.misc import create_gauss_kernel

# external libs
import numpy as np
import scipy

DR = 2  # down-sample rate


class CentroidExtractor(object):

    def __init__(self, img, cfg=None, sta=None, M=None):

        # input variables
        self._img = img
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()
        self._M = M if M is not None else self.cfg.params[self.cfg.ptc_leng]

        # private variables
        self._peak_img = self._img.copy()
        self._centroids = []

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # compute LoG to remove high frequency noise and emphasize peaks
        self.compute_log()

        # find micro image centers
        self.compute_centroids()

        # write centroids image to hard drive (if option set)
        if self.cfg.params[self.cfg.opt_dbug] and not self.sta.interrupt:
            draw_obj = CentroidDrawer(self._peak_img, self._centroids, self.cfg)
            draw_obj.write_centroids_img(fn='wht_img+mics_nms.png')
            del draw_obj

        return True

    def compute_log(self):
        """ compute Laplacian of Gaussian (LoG) """

        # print status
        self.sta.status_msg('Compute LoG', self.cfg.params[self.cfg.opt_prnt])

        # Gaussian sigma
        sig = int(self._M/4)/1.18

        # convolutions
        laplace_kernel = np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]])
        gauss_kernel = create_gauss_kernel(int(sig*6), sig)
        mexican_hat = -scipy.signal.convolve2d(gauss_kernel, laplace_kernel, 'same')
        self._peak_img = scipy.signal.convolve2d(self._img, mexican_hat, 'same')

        # print progress
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def compute_centroids(self):
        """ find coordinates of micro image centers """

        # compute local maxima with non-maximum suppression (NMS) using Down-sampling Rate (DR)
        nms_obj = NonMaxSuppression(self._peak_img[::DR, ::DR], self.cfg, self.sta)
        nms_obj.main()
        max_idx = nms_obj.idx * DR   # multiply by DR to compensate for index
        del nms_obj

        # remove indices at image borders and create list of valid integer centroid values
        h, w = self._peak_img.shape[:2]
        r = int(self._M/2)-1  # pixel distance from image border telling which maxima are excluded
        valid_idx = np.where((max_idx[0] > r) & (max_idx[1] > r) & (h-max_idx[0] > r) & (w-max_idx[1] > r))[0]
        self._centroids = list(zip(max_idx[0][valid_idx], max_idx[1][valid_idx]))

        # remove centroids that lie below threshold
        thresh = np.mean(self._peak_img)
        valid_idx = self._peak_img[np.asarray(self._centroids)[:, 0], np.asarray(self._centroids)[:, 1]] > thresh
        self._centroids = list(np.array(self._centroids)[valid_idx])

        return True

    @property
    def centroids(self):
        return self._centroids

    @property
    def peak_img(self):
        return self._peak_img



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
    from tkinter.filedialog import askopenfilename, askdirectory
except ImportError:
    import Tkinter as tk
    from tkFileDialog import askopenfilename, askdirectory

import os

from plenopticam.gui.constants import PX, PY, BTN_W


# make object for plot widget
class PathWidget(tk.Frame):

    def __init__(self, parent, path='', title="Browse", path_type=False, file_exts=()):

        # inheritance
        tk.Frame.__init__(self, parent)
        self.parent = parent

        # store extension names and types
        self.title = title
        self.path_type = path_type
        self.file_exts = file_exts

        # create current path and entry placeholder
        self._path = tk.StringVar(value=path)
        self._path_observers = []
        self.ent = tk.Entry(self, width=2*BTN_W, textvariable=self._path)
        self.ent.grid(row=0, column=0, padx=PX, pady=PY)
        self.ent.xview_moveto(1.0)  # display path from most right

        # create search button
        self.btn = tk.Button(self, width=BTN_W, text=self.title, command=self.choose_path)
        self.btn.grid(row=0, column=1, padx=PX, pady=PY)

    def choose_path(self):
        """ choose a path on the system using tkinter's frames """

        # select path considering type option (file=True; directory=False)
        if self.path_type:
            new_path = askdirectory(title=self.title, initialdir=os.path.dirname(self._path.get()),
                                    mustexist=True)
        else:
            new_path = askopenfilename(title=self.title,
                                       initialdir=os.path.dirname(self._path.get()),
                                       initialfile=os.path.basename(self._path.get()),
                                       filetypes=self.file_exts)

        # update entry text with the new path
        if new_path:
            self.path = new_path

    # enable observer pattern to track change in self.path
    @property
    def path(self):
        return self._path.get()

    @path.setter
    def path(self, val):
        self._path.set(val)
        for callback in self._path_observers:
            callback(self._path.get())

    def bind_to(self, callback):
        self._path_observers.append(callback)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import create_gauss_kernel
from plenopticam.misc.status import PlenopticamStatus

# external libs
import numpy as np
import scipy


class PitchEstimator(object):

    def __init__(self, img, cfg=None, sta=None, scale_val=1.625, precision=10, CR=3):

        # input variables
        self._img = img
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()
        self._scale_val = scale_val
        self._precision = precision
        self._CR = CR

        # internal variables
        self._top_img = None

        # output variables
        self._M = None
        self.scale_space = []

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Estimate micro image size', self.cfg.params[self.cfg.opt_prnt])

        # create scale space
        self.create_scale_space()

        # find first maximum in scale space
        self.find_scale_max(rel_max_opt=True)

        # print "Progress: Done!"
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        return True

    def create_scale_space(self):

        # take fractional central part of image
        self._crop_img(CR=self._CR)

        # create Gaussian kernels with sigma=1 and sigma=sqrt(2)
        sig_one_kernel = create_gauss_kernel(length=9, sigma=1.)
        sig_sqrt_kernel = 1 * create_gauss_kernel(length=9, sigma=np.sqrt(2))

        # initialize scale space variables
        self.scale_space = []
        img_scale = scipy.signal.convolve2d(self._top_img, sig_one_kernel, 'same')   # initial scale

        # compute scale space by Gaussian filtering and down-sampling
        while (img_scale.shape[0] or img_scale.shape[1]) >= 3:

            # filter scaled image by Gaussian kernel with sigma=1
            gauss_img_one = (scipy.signal.convolve2d(img_scale, sig_one_kernel, 'same'))

            # append scaled image to pyramid
            self.scale_space.append(-(gauss_img_one - img_scale))  # negative for maximum detection

            # filter scaled image by Gaussian kernel with sigma=sqrt(2)
            gauss_img_two = (scipy.signal.convolve2d(gauss_img_one, sig_sqrt_kernel, 'same'))

            # append scaled image to pyramid
            self.scale_space.append(-(gauss_img_two - gauss_img_one))  # negative for maximum detection

            # down-sample to half the image resolution where Gaussian filters prevent from aliasing
            img_scale = gauss_img_two[::2, ::2]

            # check interrupt status
            if self.sta.interrupt:
                return False

        return True

    def _crop_img(self, CR=1):

        # crop image (for computational speed-up and smaller region of interest)
        S = self._img.shape[0]//2
        self._top_img = self._img[S-S//CR:S+S//CR-1, S-S//CR:S+S//CR-1].copy().astype(np.float64)

        # darken image edges to exclude micro images at border
        self._top_img *= create_gauss_kernel(length=S // CR * 2 - 1, sigma=S // CR // 2)

        return True

    def get_maxima(self):
        """ compute list of maxima in scale space """

        return list(map(lambda x: np.max(x), self.scale_space))

    @staticmethod
    def interpolate_maxima(maxima, precision=10):

        nu = np.arange(0, len(maxima) - 1, 1. / precision)
        i_fun = scipy.interpolate.interp1d(range(0, len(maxima)), maxima, kind='cubic')
        maxima = i_fun(nu)
        maxima /= maxima.max()

        return maxima, nu

    def find_scale_max(self, rel_max_opt=False):
        """ determine dominant scale size by analyzing maximum over scale space """

        maxima = np.array(self.get_maxima())

        if len(maxima) > 2:
            # find scale maximum (with sub-scale precision using interpolation)
            maxima, nu = self.interpolate_maxima(maxima, self._precision)

            # start looking from first positive gradient to exclude early maxima on falling curve (false noisy peaks)
            start = np.argmax(-1*np.sign(np.gradient(nu, int(self._precision))))-1
            start = max(start, 0)

            # compute global and relative maxima in scale space
            val_max, arg_max = np.max(maxima[start:]), nu[np.argmax(maxima[start:])+start]
            rel_max, arg_rel_max = maxima[min(scipy.signal.argrelmax(maxima[start:])[0])+start], \
                                   nu[min(scipy.signal.argrelmax(maxima[start:])[0])+start]

            if rel_max_opt:
                # use global max only if its y/x ratio is larger than that of relative to penalize "weak late maxima"
                scale_max = arg_max if val_max/arg_max > rel_max/arg_rel_max else arg_rel_max
            else:
                scale_max = arg_max

        elif len(maxima) > 0:
            scale_max = max(maxima)
        else:
            scale_max = 0

        # compute corresponding DoG and Laplacian sigmas
        dog_sig = 2**(int(scale_max)//2) * np.sqrt(2**np.mod(scale_max, 2))
        lap_sig = dog_sig * 1.18

        # scale to get micro image size
        self._M = int(np.round(lap_sig * 4))

        return self._M

    @property
    def M(self):
        return self._M



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.misc.errors import PlenopticamError
from plenopticam.misc.os_ops import mkdir_p
from plenopticam.misc.type_checks import *
from plenopticam.cfg.constants import PARAMS_KEYS, PARAMS_VALS, CALIBS_KEYS

# external libs
import json
from os.path import join, abspath, dirname, basename, splitext, isdir, isfile, exists, expanduser
from os import remove, stat, chmod
import warnings


class PlenopticamConfig(object):

    # static class variables for key parameters
    lfp_path, cal_path, cal_meta, cal_meth, smp_meth, \
    ptc_leng, \
    ran_refo, \
    opt_cali, opt_vign, opt_lier, opt_cont, opt_colo, opt_awb_, opt_sat_, opt_view, opt_refo, opt_refi, opt_pflu, \
    opt_arti, opt_rota, opt_dbug, opt_prnt, opt_dpth, dir_remo \
    = PARAMS_KEYS

    pat_type, ptc_mean, mic_list = CALIBS_KEYS

    def __init__(self):

        # dicts initialization
        self.params = {}
        self.calibs = {}
        self.lfpimg = {}

        self._file_name = 'cfg.json'
        self._dir_path = dirname(abspath(__file__))     # for pip installed versions
        if not isdir(self._dir_path) and isdir(join(abspath('.'), 'cfg')):
            self._dir_path = join(abspath('.'), 'cfg')  # for py2app contents

        try:
            self.read_params()
        except (PlenopticamError, FileNotFoundError):
            self.default_values()
            self.save_params()

    def read_params(self, fp=None):

        # use default path only if file path not provided
        fp = join(self._dir_path, 'cfg.json') if fp is None else fp

        try:
            with open(fp, 'r') as f:
                json_data = json.load(f)

            # transfer parameters to config object
            for key in json_data:
                self.params[key] = str2type(json_data[key])

        except FileNotFoundError:
            pass

        # test if config parameters present
        if not self.params.keys():
            raise FileNotFoundError('Config file could not be loaded')
        # number of values in loaded config is supposed to equal config constants specified in the tool
        if not len(self.params.keys()) == len(PARAMS_KEYS):
            warnings.warn('Config file corrupted', UserWarning)

        return True

    def save_params(self, fp=None):

        if not fp:
            fp = join(self._dir_path, 'cfg.json')

        try:
            # create config folder (if not already present)
            mkdir_p(self._dir_path)
            # amend write privileges of (potentially existing) config file
            if exists(fp):
                st = stat(fp)
                chmod(fp, st.st_mode | 0o111)
            # write config file
            with open(fp, 'w+') as f:
                json.dump(self.params, f, sort_keys=True, indent=4, cls=NumpyTypeEncoder)
        except PermissionError:
            warnings.warn('\n\nGrant permission to write to the config file '+fp, UserWarning)

        return True

    def default_values(self):

        # reconstruct dict from constants
        self.params = dict(zip(PARAMS_KEYS, PARAMS_VALS))

        # write to json file
        self.save_params()

        return True

    def reset_values(self):

        self.params[self.lfp_path] = ''
        self.params[self.cal_path] = ''
        self.params[self.cal_meta] = ''
        self.params[self.ran_refo] = [0, 2]
        self.params[self.opt_refi] = False
        self.params[self.opt_awb_] = True
        self.params[self.opt_sat_] = False
        self.params[self.opt_cont] = False
        self.params[self.opt_lier] = False

    def load_cal_data(self, fp=None):

        # construct file path
        fp = self.get_file_path() if fp is None else fp

        self.calibs = self.load_json(fp=fp)

        return True

    def save_cal_data(self, fp=None, **kwargs):

        # construct file path
        fp = self.get_file_path() if fp is None else fp

        # pass kwargs to calibration dictionary (override existing)
        for kw, arg in list(kwargs.items()):
            self.calibs[kw] = arg

        self.save_json(fp=fp, **self.calibs)

        return True

    def get_file_path(self):

        # construct file path
        if not self.params[self.cal_meta]:
            if isfile(self.params[self.cal_path]):
                fp = splitext(self.params[self.cal_path])[0]+'.json'
            else:
                fp = self.params[self.cal_meta]
        else:
            fp = self.params[self.cal_meta]

        return fp

    @staticmethod
    def load_json(fp=None, sta=None):

        sta = sta if sta is not None else PlenopticamStatus()

        # file name and file path handling
        if fp is not None and splitext(fp)[-1] != '.json':
            fn = splitext(basename(fp))[0]+'.json'
            fp = join(splitext(fp)[0], fn)

        # load calibration data from json file
        if exists(fp):
            try:
                with open(fp, 'r') as f:
                    json_dict = json.load(f)
            except json.decoder.JSONDecodeError:
                remove(fp)
                sta.status_msg('Calibration JSON File may be corrupted. Attempt to delete file %s' % fp, opt=True)
                raise PlenopticamError('Calibration JSON File may be corrupted. Attempt to delete file %s' % fp)
        else:
            json_dict = None
            if splitext(fp)[-1].lower() in ('.lfp', 'lfr', '.raw'):
                sta.status_msg('Provided file %s does not exist' % fp, opt=True)

        return json_dict

    @staticmethod
    def save_json(fp=None, **kwargs):

        # filename and file path handling
        if fp is not None:

            # json extension handling
            if splitext(fp)[-1] != '.json':
                fn = basename(splitext(fp)[0])+'.json'
                fp = join(dirname(splitext(fp)[0]), fn)

            # create folder
            mkdir_p(dirname(fp), False)

        # save calibration data as json file
        json_dict = kwargs['json_dict'] if 'json_dict' in kwargs else kwargs
        try:
            # amend write privileges of (potentially existing) config file
            if exists(fp):
                st = stat(fp)
                chmod(fp, st.st_mode | 0o111)
            # write file
            with open(fp, 'wt') as f:
                json.dump(json_dict, f, sort_keys=True, indent=4)
        except TypeError:
            return False

        return True

    @property
    def exp_path(self):
        """ export directory path """
        return splitext(self.params[self.lfp_path])[0] if self.params[self.lfp_path] else expanduser("~")

    def cond_load_limg(self, img=None):
        return exists(self.params[self.lfp_path]) and (img is None and self.cond_lfp_align() or self.cond_auto_find())

    def cond_auto_find(self):
        return isdir(self.params[self.cal_path]) or self.params[self.cal_path].lower().endswith('.tar')

    def cond_load_wimg(self):
        return not self.cond_auto_find() and \
               (self.params[self.opt_cali] or self.params[self.opt_vign] or self.cond_lfp_align())

    def cond_perf_cali(self):
        return self.params[self.opt_cali] or \
               (not self.cond_meta_file() and self.cond_lfp_align())

    def cond_lfp_align(self):
        return not exists(join(self.exp_path, 'lfp_img_align.pkl'))

    def cond_meta_file(self):

        # look for meta data file (optionally find json file named after calibration image file)
        pot_meta = splitext(self.params[self.cal_path])[0] + '.json'
        cal_meta = self.params[self.cal_meta]
        self.params[self.cal_meta] = pot_meta if not isfile(cal_meta) and isfile(pot_meta) else cal_meta
        exist = isfile(self.params[self.cal_meta])

        if exist:
            # load meta data file and validate content
            self.load_cal_data()
            min_res_y = max(map(lambda x: x[2], self.calibs[self.mic_list])) * self.calibs['ptc_mean'][0]
            min_res_x = max(map(lambda x: x[3], self.calibs[self.mic_list])) * self.calibs['ptc_mean'][1]
            valid = min_res_y > 0 and min_res_x > 0
        else:
            valid = False

        return exist and valid


class NumpyTypeEncoder(json.JSONEncoder):
    def default(self, obj):
        import numpy as np
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyTypeEncoder, self).default(obj)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os, time, datetime
from plenopticam.misc.status import PlenopticamStatus


class PlenopticamError(Exception):

    URL_ISSUE = 'https://github.com/hahnec/plenopticam/issues/new'

    def __init__(self, *args, **kwargs):
        super(PlenopticamError, self).__init__(*args)

        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else None
        self.sta = kwargs['sta'] if 'sta' in kwargs else PlenopticamStatus()

        self.args = args
        try:
            self.write_log()
        except PermissionError as e:
            self.sta.status_msg(msg=e, opt=True)
            raise e

    def write_log(self):

        # log file name
        fn = 'err_log.txt'

        # export folder path
        fp = self.cfg.exp_path if hasattr(self.cfg, 'exp_path') else os.path.expanduser("~")

        # send status to user while referring to log file
        self.sta.status_msg('Error! See log file in %s.' % os.path.join(os.pardir, os.path.basename(fp), fn))

        # write error to log file
        with open(os.path.join(fp, fn), 'a') as f:
            f.writelines(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))
            f.writelines('\nOpen issue at %s and paste below traceback.\n\n' % self.URL_ISSUE)
            f.writelines(self.args.__str__())
            f.writelines('\n\n\n')


class LfpTypeError(PlenopticamError):
    def __init__(self, *args, **kwargs):
        super(LfpTypeError, self).__init__(*args, **kwargs)


class LfpAttributeError(PlenopticamError):
    def __init__(self, *args, **kwargs):
        super(LfpAttributeError, self).__init__(*args, **kwargs)

#from requests import session
#from bs4 import BeautifulSoup as bs
#
#class GithubReporter(PlenopticamError):
#
#    USER = 'PlenopticamReporter'
#    PASSWORD = 'u9?KX2+y'
#
#    def __init__(self, *args, **kwargs):
#        PlenopticamError.__init__(self, *args)
#
#        self.url = 'https://github.com/session'
#        self.url_query = self.URL_ISSUE + '?labels=auto&title=Automatic+issue+report&assignees=hahnec'
#
#        self.s = session()
#
#    def alt_submit(self):
#
#        from requests.auth import HTTPBasicAuth
#        self.s.get(self.url_query, auth=HTTPBasicAuth(self.USER, self.PASSWORD))
#
#    def logout(self):
#        pass
#
#    def login(self):
#
#        req = self.s.get(self.url).text
#        html = bs(req)
#        token = html.find("input", {"name": "authenticity_token"}).attrs['value']
#        com_val = html.find("input", {"name": "commit"}).attrs['value']
#
#        login_data = {'login': self.USER,
#                      'password': self.PASSWORD,
#                      'commit': com_val,
#                      'authenticity_token': token
#        }
#
#        self.s.post(self.url, data=login_data)
#
#        return True
#
#    def issue(self):
#
#        site = self.s.get(self.url_query).text
#        html = bs(site)
#
#        token = html.find("input", {"name": "authenticity_token"}).attrs['value']
#        #com_val = html.find("button", {"name": "commit"}).attrs['value']
#
#        issue_data = {
#            'issue': [{
#                "title": 'Plenopticam Error Report',
#                "body": 'Test'
#            }]
#            #'issue[body]': 'Test'#,
#            #'commit': com_val,
#            #'authenticity_token': token
#        }
#
#        self.s.post(self.url_query, data=issue_data)
#
#    def __del__(self):
#
#        # logout from helper account
#        self.logout()
#
#        # close session
#        self.s.close()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from PIL import Image, ImageFont, ImageDraw
import os, sys
from functools import partial
import glob

try:
    from PIL.ImageTk import PhotoImage
except:
    PhotoImage = tk.PhotoImage

from plenopticam.lfp_extractor import LfpViewpoints
from plenopticam.misc.os_ops import get_img_list
from plenopticam import __version__
from plenopticam.gui.constants import PX, PY


class ViewWidget(tk.Canvas, LfpViewpoints):

    def __init__(self, *args, **kwargs):

        # use button status for enabling/disabling purposes
        self.btn = kwargs['btn'] if 'btn' in kwargs else {'state': 'normal'}
        kwargs.pop('btn') if 'btn' in kwargs else None
        self.btn['state'] = tk.DISABLED

        LfpViewpoints.__init__(self, *args, **kwargs)
        kwargs.pop('cfg') if 'cfg' in kwargs else None
        kwargs.pop('sta') if 'sta' in kwargs else None

        tk.Canvas.__init__(self, *args, **kwargs)

        # take images from arguments or load from hard drive
        if 'vp_img_arr' in kwargs and 'refo_stack' in kwargs:
            self.vp_img_arr = kwargs['vp_img_arr']
            self.refo_stack = kwargs['refo_stack']
        else:
            self.load_data()

        # window settings
        self.master.title("PlenoptiCam Viewer")
        self.set_ascii_symbols()
        self.shape = self.vp_img_arr.shape[2:] if self.vp_img_arr is not None else (PY*30, PX*40, 1)
        self._ht = self.shape[0]*2 + PY*16
        self._wd = self.shape[1]*2 + PX*16

        # initialize member variables
        self.vp_mode = True
        self.auto_mode = False
        self._mode_text = tk.StringVar()
        self._mode_text.set('V' if self.vp_mode else 'R')
        self.all_function_trigger()

        # display initial image
        self.show_image()

        # start with auto play mode
        self.auto_play()

    def load_data(self):
        """ automatically look for output folders and load light field content """

        # load list of potential directories
        vp_dirs = glob.glob(os.path.join(self.cfg.exp_path, 'viewpoints_*px'))
        rf_dirs = glob.glob(os.path.join(self.cfg.exp_path, 'refo_*px'))

        # exclude up-sampled refocused images (due to oversize)
        terms = ['upscale_']
        rf_dirs = [dir for substr in terms for dir in rf_dirs if not os.path.basename(dir).__contains__(substr)]

        # load data
        self.vp_img_arr = self.select_from_dir(vp_dirs, vp=1)
        self.refo_stack = self.select_from_dir(rf_dirs, vp=0)

        # obtain micro image size of loaded data
        self._size_pitch = self.vp_img_arr.shape[0] if self.vp_img_arr is not None else self._size_pitch

        # set images to None if desired micro image size is bigger than in loaded data
        if self._size_pitch < self.cfg.params[self.cfg.ptc_leng]:
            self.vp_img_arr = None
            self.refo_stack = None

        # light-field related data
        self.reset_indices()
        if self.vp_img_arr is not None:
            r = int(self.cfg.params[self.cfg.ptc_leng]//2)
            self.move_coords = self.get_move_coords(arr_dims=self.vp_img_arr.shape[:2], pattern='circle', r=r)

        return True

    def select_from_dir(self, dirs, vp=None):

        # select current micro image size
        size_idx = 0
        for i, s in enumerate(dirs):
            size = int(s.split('_')[-1][:-2])
            if size >= self.cfg.params[self.cfg.ptc_leng]:
                size_idx = i
                break

        # load image list
        try:
            ret_val = get_img_list(dirs[size_idx], vp=vp)
        except IndexError:
            ret_val = None

        return ret_val

    def show_image(self, arg=None):

        # compute new light-field positions depending on pressed button
        self._adapt_coords(arg)

        # set new frame considering view/refo mode
        if self.vp_mode and self.vp_img_arr is not None:
            next_frame = Image.fromarray(self.vp_img_arr[self._v, self._u, ...])
        elif self.refo_stack is not None:
            next_frame = Image.fromarray(self.refo_stack[self._a])
        else:
            # display text message about missing image data
            next_frame = self.get_dummy

        # set tk frame as member variable as it gets lost otherwise
        self.tk_frame = PhotoImage(next_frame)

        self.delete(self.find_withtag("bacl"))
        self.allready = self.create_image(PX*4, PY*4, image=self.tk_frame, anchor=tk.NW, tag="bacl")
        self.find_withtag("bacl")

    def _adapt_coords(self, arg=None):

        arg = -1 if arg is None else arg

        # consider which button was pressed
        if self.vp_mode:
            if arg == 0:
                self._u += 1 if self._u < self._size_pitch-1 else 0
            if arg == 1:
                self._u -= 1 if self._u > 0 else 0
            if arg == 2:
                self._v += 1 if self._v < self._size_pitch-1 else 0
            if arg == 3:
                self._v -= 1 if self._v > 0 else 0
        else:
            if arg == 0:
                self._a += 1 if self._a < len(self.refo_stack)-1 else 0
            if arg == 1:
                self._a -= 1 if self._a > 0 else 0

        return True

    def all_function_trigger(self):

        self.create_buttons()
        self.window_settings()

        return True

    def window_settings(self):

        self['width'] = int(self._wd/2)
        self['height'] = int(self._ht/2)

        return True

    def create_buttons(self):

        # auto-play button
        self._btn_auto_text = tk.StringVar()
        self._btn_auto_text.set(self._loop_symbol)
        self.btn_auto = tk.Button(self, textvariable=self._btn_auto_text, command=self.auto_play, height=1, width=2)
        self.btn_auto.place(x=PX, y=PY, anchor=tk.NW)

        # mode button
        self.btn_mode = tk.Button(self, textvariable=self._mode_text, command=self.switch_mode, height=1, width=2)
        self.btn_mode.place(x=PX, y=PY*4, anchor=tk.NW)

        # instantiate button objects for light-field
        self.btn_arrows = list()
        btn_pos = [[self.shape[1]+PX*4.5, PX, self._wd/4, self._wd/4],
                   [self._ht/4, self._ht/4, PY, self.shape[0]+PY*4.5]]
        for i, text in enumerate(self._arrow_symbols):
            next_frame_arg = partial(self.show_image, i)
            self.btn_arrows.append(tk.Button(self, text=text, command=next_frame_arg, height=1, width=2))
            self.btn_arrows[i].place(x=btn_pos[0][i], y=btn_pos[1][i], anchor=tk.NW)

        return True

    def switch_mode(self):

        # switch mode
        self.vp_mode = not self.vp_mode

        # disable buttons in 1-D mode
        for idx in range(2, 4):
            self.btn_arrows[idx]['state'] = tk.NORMAL if tk.DISABLED else tk.DISABLED

        # alternate button text
        self._mode_text.set('V' if self.vp_mode else 'R')
        self.btn_mode.config(textvariable=self._mode_text)

        # reset indices
        self.reset_indices()

        # load image
        self.show_image()

    def reset_indices(self):
        """ reset light-field indices to default """

        self._u, self._v, self._a, self._k, self._h = (self._size_pitch // 2 + 1, self._size_pitch // 2 + 1, -1, 0, 1)

    def set_ascii_symbols(self):
        """ load individual ASCII symbols for buttons as they appear different on each OS """

        if sys.platform.startswith('dar'):
            # macOS
            self._pause_symbol = '❙❙'
            self._loop_symbol = '↻'
            self._arrow_symbols = [' ▶ ', ' ◀ ', ' ▲ ', ' ▼ ']
        elif sys.platform.startswith('lin'):
            # UNIX
            self._pause_symbol = '❙❙'
            self._loop_symbol = '↻'
            self._arrow_symbols = [' ▶ ', ' ◀ ', ' ▲ ', ' ▼ ']
        elif sys.platform.startswith('win'):
            # Windows
            self._pause_symbol = '⏸'
            self._loop_symbol = '⏯'
            self._arrow_symbols = [' ⏵ ', ' ⏴ ', ' ⏶ ', ' ⏷ ']
        else:
            # better than nothing
            self._pause_symbol = '⏸'
            self._loop_symbol = '⏯'
            self._arrow_symbols = [' ⏵ ', ' ⏴ ', ' ⏶ ', ' ⏷ ']

    def auto_play(self):

        self.auto_mode = not self.auto_mode

        if self.auto_mode:
            self._btn_auto_text.set(self._pause_symbol)
        else:
            self._btn_auto_text.set(self._loop_symbol)

        self.auto_loop()

    def auto_loop(self):

        if self.auto_mode:

            if self.vp_mode and self.vp_img_arr is not None:

                # use k index for coordinate iteration of viewpoints
                self._k = 0 if self._k >= len(self.move_coords) else self._k
                self._u, self._v = self.move_coords[self._k]

                # wait some time before showing next image
                self.after(100, self.auto_loop)
                self.show_image()

            elif self.refo_stack is not None:

                # use k and h indices for forward and backward iteration of refocusing slice index a
                self._k, self._h = (0, -self._h) if self._k >= len(self.refo_stack) else (self._k, self._h)
                self._a = self._k*self._h if self._k != 0 or self._h > 0 else -1

                # wait some time before showing next image
                self.after(500, self.auto_loop)
                self.show_image()

            self._k += 1

    def destroy(self):
        """ reset button state for close window event """

        # prevent loop iteration from running after window closure
        self.auto_mode = False

        # reset button status once window is closed
        self.btn['state'] = tk.NORMAL

    @property
    def get_dummy(self):
        """ notification image about data availability """

        # disable auto play mode
        self.auto_mode = False
        self._btn_auto_text.set(self._loop_symbol)

        # create notification image
        text = "%s image data not found" % ('Viewpoint' if self.vp_mode else 'Refocused')
        font = ImageFont.load_default()
        img = Image.new("RGBA", self.shape[:2][::-1], color=(0, 0, 0, 0))
        draw = ImageDraw.Draw(img)
        w, h = draw.textsize(text, font)
        draw.text(((self.shape[1] - w) / 2, (self.shape[0] - h) / 2), text, fill=(0, 0, 0), font=font)
        ImageDraw.Draw(img)

        return img


def main():

    # creating window
    root = tk.Tk(className="PlenoptiCam Viewer "+__version__)
    # creating canvas widget
    ViewWidget(root).pack(expand="no", fill="both")
    # make not resizable
    root.resizable(width=0, height=0)
    # window mainloop
    root.mainloop()

    return True


if __name__ == '__main__':

    main()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import os
from os.path import abspath, dirname, basename
import requests
from zipfile import ZipFile

from plenopticam.misc import mkdir_p, PlenopticamStatus
from plenopticam.cfg import PlenopticamConfig


class DataDownloader(object):

    def __init__(self, *args, **kwargs):

        # instantiate config and status objects
        self.cfg = kwargs['cfg'] if 'cfg' in kwargs else PlenopticamConfig()
        self.sta = kwargs['sta'] if 'sta' in kwargs else PlenopticamStatus()

        # path handling: refer to folder where data will be stored
        path = kwargs['path'] if 'path' in kwargs else os.path.dirname(os.path.abspath(__file__))
        self.fp = os.path.join(path, 'data')
        self.root_path = dirname(abspath('.')) if basename((abspath('.'))) == 'tests' else abspath('.')

        # data urls
        self.host_eu_url = 'http://wp12283669.server-he.de/Xchange/Binder/illum_test_data.zip'
        self.opex_prep_url = 'http://wp12283669.server-he.de/Xchange/Binder/OpEx_data_prepared.zip'
        self.opex_url = 'https://ndownloader.figshare.com/files/5201452'
        self.opex_fnames_wht = ['f197with4m11pxf16Final.bmp', 'f197Inf9pxFinalShift12.7cmf22.bmp']
        self.opex_fnames_lfp = ['f197with4m11pxFinal.bmp', 'f197Inf9pxFinalShift12.7cm.bmp']

    def download_data(self, url, fp=None):
        """ download data form provided url string """

        # path handling
        self.fp = fp if fp is not None else self.fp
        mkdir_p(self.fp) if not os.path.exists(self.fp) else None

        # skip download if file exists
        if os.path.exists(os.path.join(self.fp, os.path.basename(url))):
            print('Download skipped as %s already exists' % os.path.basename(url))
            return None

        print('Downloading file %s to %s' % (os.path.basename(url), self.fp))

        with open(os.path.join(self.fp, os.path.basename(url)), 'wb') as f:
            # establish internet connection for data download
            try:
                r = requests.get(url, stream=True)
                total_length = r.headers.get('content-length')
            except requests.exceptions.ConnectionError:
                raise Exception('Check your internet connection, which is required for downloading test data.')

            if total_length is None:  # no content length header
                f.write(r.content)
            else:
                dl = 0
                total_length = int(total_length)
                for data in r.iter_content(chunk_size=4096):
                    dl += len(data)
                    f.write(data)
                    self.sta.progress(dl/total_length * 100, self.cfg.params[self.cfg.opt_prnt])

        print('\n Finished download of %s' % os.path.basename(url))

    def extract_archive(self, archive_fn=None, fname_list=None, fp=None):
        """ extract content from downloaded data """

        # look for archives in file path
        self.fp = fp if fp is not None else self.fp
        if archive_fn is None and fp:
            archive_fns = [os.path.join(self.fp, f) for f in os.listdir(self.fp) if f.endswith('zip')]
        else:
            archive_fns = [archive_fn]

        for archive_fn in archive_fns:

            # choose from filenames inside archive
            fname_list = ZipFile(archive_fn).namelist() if fname_list is None else fname_list
            fname_list = self.find_archive_fnames(archive_fn, 'caldata', 'lfr') if fname_list == 'lfr' else fname_list

            # extract chosen files
            with ZipFile(archive_fn) as z:
                for fn in z.namelist():
                    if fn in fname_list and not os.path.exists(os.path.join(self.fp, fn)):
                        z.extract(fn, os.path.dirname(archive_fn))
                        print('Extracted file %s' % fn)

    @staticmethod
    def find_archive_fnames(archive_fn, head_str='', tail_str=''):
        return [f for f in ZipFile(archive_fn).namelist() if f.startswith(head_str) or f.endswith(tail_str)]

    @property
    def fp(self):
        return self._fp

    @fp.setter
    def fp(self, fp):
        self._fp = fp



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2021 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import unittest

import numpy as np
from os.path import join
import zipfile
from scipy.spatial.distance import cdist

from plenopticam.lfp_calibrator import CentroidSorter, GridFitter, CentroidFitSorter, find_centroid
from plenopticam.cfg import PlenopticamConfig, constants
from plenopticam.misc import load_img_file


class PlenoptiCamTesterCalib(unittest.TestCase):

    CEA_PATH = r'../examples/data/synth_spots'

    def __init__(self, *args, **kwargs):
        super(PlenoptiCamTesterCalib, self).__init__(*args, **kwargs)

        self.PLOT_OPT = False

    def setUp(self):

        # instantiate config and status objects
        self.cfg = PlenopticamConfig()
        self.cfg.default_values()

        # omit status messages
        self.cfg.params[self.cfg.opt_prnt] = False

        # enable options in config to cover more algorithms in tests
        self.cfg.params[self.cfg.cal_meth] = constants.CALI_METH[3]

        try:
            with zipfile.ZipFile(self.CEA_PATH+'.zip', 'r') as zip_obj:
                zip_obj.extractall(self.CEA_PATH)
        except BaseException:
            pass

    def test_mla_geometry_estimate(self):

        pat_types = ['rec', 'hex']
        max_dim = 30
        pitch = 1

        for flip in [False, True]:
            for pat_type in pat_types:
                pseudo_grid = GridFitter.grid_gen(dims=[max_dim, max_dim], pat_type=pat_type, hex_odd=False)
                pseudo_grid = pseudo_grid[:, :2][:, ::-1] if flip else pseudo_grid
                sorter = CentroidSorter(centroids=pseudo_grid)
                pattern = sorter._estimate_mla_geometry(pitch=pitch)
                self.assertEqual(pat_type, pattern, 'Pattern detection failed')

        return True

    def test_grid_gen(self):

        dim = 5
        x = np.linspace(-.5, .5, dim)
        i = np.linspace(0, dim-1, dim)
        coords = np.concatenate([np.meshgrid(x.T, x)[::-1], np.meshgrid(i.T, i)[::-1]], axis=0).reshape(-1, dim**2).T

        pseudo_grid = GridFitter.grid_gen(dims=[dim, dim], pat_type='rec', hex_odd=False, normalize=True)

        rmse = np.sqrt(np.sum(np.square(coords-pseudo_grid)))

        self.assertEqual(rmse, 0, 'Grid generation failed')

    def test_grid_rotation_fit(self):

        # test grid generation
        dims = [122, 122]
        pat_type = 'rec'
        grid = GridFitter.grid_gen(dims=dims, pat_type=pat_type)

        # parameters for regression
        rvecs = [15*np.random.rand(3) for _ in range(5)]
        compose, affine, flip_xy = [False] * 3
        compose = False

        for rvec in rvecs:

            # create transformation matrix
            rvec = rvec/180*np.pi
            rmat = GridFitter.euler2mat(*rvec)
            rmat[-1, -1] = 1

            # apply transformation
            xpts = GridFitter().apply_transform(rmat, grid.copy(), affine, flip_xy, z_dist=np.max(grid[:, :2]))

            # grid regression
            gf = GridFitter(xpts, compose=compose, affine=affine, flip_xy=flip_xy, z_dist=np.max(grid[:, :2]))
            gf.main()

            # rotation matrix assertion
            all_close = np.allclose(gf.pmat, rmat, rtol=1e-01, atol=1e-01)
            self.assertTrue(all_close, 'Grid estimation failed')

            # decomposition matrix assertion
            xrmat = gf.decompose(gf.pmat)[0]
            all_close = np.allclose(rmat, xrmat, rtol=1e-01, atol=1e-01)
            self.assertTrue(all_close, 'QR-decomposition failed')

            # euler angles assertion
            angles = gf.mat2euler(xrmat)
            all_close = np.allclose(angles, rvec)
            self.assertTrue(all_close, 'Angle estimation failed')

    def test_decomposition(self):

        for _ in range(100):

            rvec = np.random.randn(3)/180*np.pi
            rmat = GridFitter.euler2mat(*rvec)
            kmat = np.diag(np.array([*np.random.rand(2)*2 + 1, 1]))
            pmat = np.dot(rmat, kmat)

            xvec = GridFitter.mat2euler(rmat)
            all_close = np.allclose(rvec, xvec)
            self.assertTrue(all_close, 'Angle extraction failed')

            xrmat, xkmat, xtvec = GridFitter.decompose(pmat)
            all_close = np.allclose([rmat, kmat], [xrmat, xkmat])
            self.assertTrue(all_close, 'Decomposition failed')

            xvec = GridFitter.mat2euler(xrmat)
            all_close = np.allclose(rvec, xvec)
            self.assertTrue(all_close, 'Angle decomposition failed')

    def test_sorted_fitting(self):

        np.random.seed(32)
        # parameter init
        data = [[(16, 23), 'hex', False, 10, 0], [(88, 88), 'hex', True, 15, 1e-6], [(9, 9), 'rec', False, 2, 1e-1]]

        for (dim_y, dim_x), pat_type, odd, num_miss, err_init in data:

            # generation of ground-truth test grid
            ground_mics = GridFitter.grid_gen(dims=[dim_y, dim_x], pat_type=pat_type, hex_odd=odd)
            ground_mics[:, :2] = ground_mics[:, :2] - np.array([ground_mics[:, 0].min(), ground_mics[:, 1].min()])

            # disturb grid
            remove_idxs = np.random.randint(dim_x, ground_mics.shape[0]-dim_x, size=num_miss, dtype=np.uint16)
            remove_idxs = np.append(remove_idxs, -10)
            obtain_mics = np.delete(ground_mics[:, :2], remove_idxs, axis=0)
            obtain_mics[:, :2] += err_init*np.random.rand(obtain_mics.shape[0], 2)
            self.assertEqual(obtain_mics.shape[0], dim_y*dim_x-num_miss-1, msg='Unexpected number of missing centroids')

            # call 4-corner fit
            sorter = CentroidFitSorter(centroids=obtain_mics)
            sorted_mics = sorter.corner_fit()
            sorted_mics = np.array(sorted_mics)

            if self.PLOT_OPT:
                import matplotlib.pyplot as plt
                plt.plot(ground_mics.T[1], ground_mics.T[0], 'k.')
                plt.plot(obtain_mics.T[1], obtain_mics.T[0], 'r*')
                plt.plot(sorter.corner_mics.T[1], sorter.corner_mics.T[0], 'bx')
                plt.plot(sorted_mics.T[1], sorted_mics.T[0], 'g+')
                plt.show()

            # validation
            dres = cdist(ground_mics[:, :2], sorted_mics[:, :2], 'euclidean')
            idxs = np.argmin(dres, axis=0)
            rval = np.allclose(sorted_mics[:, :2], ground_mics[idxs, :2], atol=10e7)
            self.assertTrue(rval)

    def test_mla_dims_estimate(self):

        # init
        dim_y, dim_x = 9, 9
        pat_type = 'hex'
        flip = False

        # grid generation and modification
        pseudo_mics = GridFitter.grid_gen(dims=[dim_y, dim_x], pat_type=pat_type, hex_odd=False)[:, :2]
        ty, tx = np.min(pseudo_mics[:, 0]), np.min(pseudo_mics[:, 1])
        pseudo_mics -= np.array([ty, tx])
        pseudo_mics = pseudo_mics[:, ::-1] if flip else pseudo_mics

        # estimate MLA dimensions
        sorter = CentroidSorter(centroids=pseudo_mics)
        sorter._mla_dims()
        est_y, est_x = sorter._lens_y_max, sorter._lens_x_max

        self.assertEqual((est_y, est_x), (dim_y, dim_x), 'Centroid number estimation failed')

    def test_pitch_estimator(self):

        from plenopticam.lfp_calibrator import PitchEstimator

        fns = [join(self.CEA_PATH, fn+'.png') for fn in ['a', 'b', 'c', 'd']]
        ref_sizes = [141, 52, 18, 6]

        for fn, ref_size in zip(fns, ref_sizes):
            img = load_img_file(fn)
            obj = PitchEstimator(img=img, cfg=self.cfg)
            obj.main()

            self.assertEqual(ref_size, obj.M)

    def test_all(self):

        self.test_mla_geometry_estimate()
        self.test_grid_gen()
        self.test_mla_dims_estimate()
        self.test_sorted_fitting()


if __name__ == '__main__':
    unittest.main()



from .centroid_fit_sort import CentroidFitSorter
from .centroid_refiner import CentroidRefiner
from .centroid_sorter import CentroidSorter
from .centroid_extractor import CentroidExtractor
from .find_centroid import find_centroid
from .pitch_estimator import PitchEstimator
from .non_max_supp import NonMaxSuppression
from .cali_finder import CaliFinder
from .grid_fitter import GridFitter
from .line_fitter import LineFitter
from .cali_finder import CaliFinder
from .centroid_drawer import CentroidDrawer
from .top_level import LfpCalibrator

# Downsample rate for image processing speed-up
DR = 4


# local imports
from plenopticam import misc
from plenopticam.cfg import constants as c
from plenopticam.lfp_aligner.lfp_global_resampler import LfpGlobalResampler
from plenopticam.lfp_aligner.lfp_local_resampler import LfpLocalResampler

# external libs
import os
import pickle


class LfpResampler(LfpLocalResampler, LfpGlobalResampler):

    def __init__(self, *args, **kwargs):
        super(LfpResampler, self).__init__(*args, **kwargs)

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Light-field alignment', self.cfg.params[self.cfg.opt_prnt])

        # global resampling
        if self.cfg.params[self.cfg.smp_meth] == 'global' or not self.cfg.params[self.cfg.smp_meth]:
            try:
                self.global_resampling()
            except ImportError:
                self.cfg.params[self.cfg.smp_meth] = 'local'
                self.sta.status_msg('Use local resampling due to ImportError', self.cfg.params[self.cfg.opt_prnt])

        # local resampling
        if self.cfg.params[self.cfg.smp_meth] == 'local':
            self.local_resampling()

        # unrecognized resampling method
        if not self.cfg.params[self.cfg.smp_meth] in c.SMPL_METH and self.cfg.params[self.cfg.smp_meth]:
            self.sta.status_msg('Resampling method %s unrecognized.' % self.cfg.params[self.cfg.smp_meth],
                                self.cfg.params[self.cfg.opt_prnt])
            self.sta.interrupt = True

        # save aligned image to hard drive
        self._write_lfp_align()

        return True

    def _write_lfp_align(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Save aligned light-field', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # convert to 16bit unsigned integer
        self._lfp_img_align = misc.Normalizer(self._lfp_img_align).uint16_norm()

        # create output data folder
        misc.mkdir_p(self.cfg.exp_path, self.cfg.params[self.cfg.opt_prnt])

        try:
            # write aligned light field as pickle file to avoid re-calculation
            with open(os.path.join(self.cfg.exp_path, 'lfp_img_align.pkl'), 'wb') as f:
                pickle.dump(self._lfp_img_align, f)
        except pickle.UnpicklingError:
            # print status and interrupt process
            fname = os.path.join(self.cfg.exp_path, 'lfp_img_align.pkl')
            self.sta.status_msg('Pickle file may be corrupted %s' % fname, self.cfg.params[self.cfg.opt_prnt])
            self.sta.error = True

        if self.cfg.params[self.cfg.opt_dbug]:
            misc.save_img_file(self._lfp_img_align, os.path.join(self.cfg.exp_path, 'lfp_img_align.tiff'))

        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from plenopticam.gui.constants import PX, PY, BTN_W
from plenopticam.cfg import constants as c


class CnfgWidget(tk.Toplevel):

    # hidden parameter options in user interface
    EXCLUDED_PARAMS = ('opt_view', 'opt_refo', 'opt_prnt', 'opt_rota', 'opt_dbug', 'opt_colo', 'opt_lier')

    # generate settings properties in dict consisting of lists containing names and types
    PROPERTIES = dict(zip(c.PARAMS_KEYS, (pair for pair in zip(c.PARAMS_NAME, c.PARAMS_TYPE))))

    def __init__(self, parent):

        # inheritance
        tk.Toplevel.__init__(self, parent)
        self.cfg = parent.cfg

        # open window
        self.wm_title("Configuration")

        # tk variables init
        self.tk_vars = {}

        # tk widget list init
        self.obj_ents = {}

        # hide some config keys in user interface which are given as tuple
        self.gui_keys = [key for key in self.PROPERTIES.keys() if key not in self.EXCLUDED_PARAMS]

        # place properties in tk frame
        for i, key in enumerate(self.gui_keys):
            tk.Label(self, text=self.PROPERTIES[key][0]).grid(row=i, column=0, sticky='W')
            obj_ent = None
            if self.PROPERTIES[key][1] == 'str':
                self.tk_vars[key] = tk.StringVar(value=self.cfg.params[key])
                obj_ent = tk.Entry(self, textvariable=self.tk_vars[key], width=2*PX)

            elif self.PROPERTIES[key][1] == 'int':
                self.tk_vars[key] = tk.IntVar(value=int(self.cfg.params[key]))
                obj_ent = tk.Entry(self, textvariable=self.tk_vars[key], width=PX)

            elif self.PROPERTIES[key][1] == 'list':
                self.tk_vars[key] = tk.StringVar(value=str(self.cfg.params[key]))
                obj_ent = tk.Entry(self, textvariable=self.tk_vars[key], width=PX)

            elif self.PROPERTIES[key][1] == 'ran':
                self.tk_vars[key] = TwoStringVars(values=self.cfg.params[key])
                obj_ent = DoubleSpinbox(self, textvariable=self.tk_vars[key], width=PX)

            elif self.PROPERTIES[key][1] == 'sel':
                # default values
                value_ran, value_sel, scale_fld = '', '', 1
                # load value range
                if key == self.cfg.cal_meth:
                    value_ran, default = (c.CALI_METH, c.CALI_METH[0])
                    value_sel = self.cfg.params[self.cfg.cal_meth] if self.cfg.cal_meth in self.cfg.params else default
                    scale_fld = 2
                if key == self.cfg.ptc_leng:
                    value_ran, default = (c.PTCH_SIZE, c.PTCH_SIZE[2])
                    value_sel = self.cfg.params[self.cfg.ptc_leng] if self.cfg.ptc_leng in self.cfg.params else default
                if key == self.cfg.smp_meth:
                    value_ran, default = (c.SMPL_METH, c.SMPL_METH[0])
                    value_sel = self.cfg.params[self.cfg.smp_meth] if self.cfg.smp_meth in self.cfg.params else default
                self.tk_vars[key] = tk.StringVar(value=value_sel)
                obj_ent = tk.Spinbox(self, values=value_ran, textvariable=self.tk_vars[key], width=PX//2*scale_fld)
                self.tk_vars[key].set(value=value_sel)   # set to default necessary for tkinter's spinbox

            elif self.PROPERTIES[key][1] == 'bool':
                self.tk_vars[key] = tk.BooleanVar(value=bool(self.cfg.params[key]))
                obj_ent = tk.Checkbutton(self, variable=self.tk_vars[key])

            # place entry
            obj_ent.grid(row=i, column=1, sticky='W')

            # display text from most right
            obj_ent.xview_moveto(1.0) if self.PROPERTIES[key][1] != 'bool' else None

            self.obj_ents[key] = obj_ent

        self.btn = tk.Button(self, text="Save & Close", width=BTN_W*2, command=self.close)
        self.btn.grid(row=len(self.gui_keys)+1, columnspan=2, padx=PX, pady=PY)

        # makes sure no mouse or keyboard events are sent to other windows (avoid multiple simultaneous instances)
        self.grab_set()

    def refi(self):

        # set refocus refinement option if scheimpflug focus type is desired
        if self.tk_vars[self.cfg.opt_pflu].get():
            self.tk_vars[self.cfg.opt_refi].set(value=True)
            self.obj_ents[self.cfg.opt_refi].select()
            self.cfg.params[self.cfg.opt_refi] = True

    def save_cfg(self):

        # read from tkinter variables
        for key in self.gui_keys:
            self.cfg.params[key] = self.tk_vars[key].get()
            # set refinement option if scheimpflug option set
            if key == self.cfg.opt_pflu:
                self.refi()

        # write current configuration settings to hard drive
        self.cfg.save_params()

        # load parameters to current config file
        self.cfg.read_params()

    def close(self):
        """ close about window """

        # save configuration settings
        self.save_cfg()

        # destruct frame object
        self.destroy()

        return True


class TwoStringVars(tk.StringVar):

    def __init__(self, master=None, values=('', '')):
        tk.StringVar.__init__(self, master, values)

        self.set(values)

    def get(self):
        one = int(self._one.get())
        two = int(self._two.get())
        return [one, two]

    def set(self, values):
        if len(values) == 2:
            self._one = tk.StringVar(value=str(values[0]))
            self._two = tk.StringVar(value=str(values[1]))
        else:
            raise ValueError('Pass list or tuple of two values only')

    @property
    def one(self):
        return int(self._one.get())

    @property
    def two(self):
        return int(self._two.get())


class DoubleSpinbox(tk.Frame):

    def __init__(self, master=None, textvariable=None, _from=None, to=None, width=None):
        tk.Frame.__init__(self, master)

        self._v = TwoStringVars() if textvariable is None else textvariable
        self._f = -9 if _from is None else _from
        self._t = +9 if to is None else to
        self._width = 5 if width is None else width//2

        self._spinbox_one = tk.Spinbox(self, textvariable=self._v._one, from_=self._f, to=self._t, width=self._width)
        self._spinbox_one.grid(row=0, column=0, sticky='NSW')
        self._spinbox_two = tk.Spinbox(self, textvariable=self._v._two, from_=self._f+1, to=self._t+1, width=self._width)
        self._spinbox_two.grid(row=0, column=1, sticky='NSE')

    def xview_moveto(self, val):
        """ display text from most right """

        self._spinbox_one.xview_moveto(val)
        self._spinbox_two.xview_moveto(val)



def isfloat(x):
    try:
        float(x)
    except TypeError:
        return False
    except ValueError:
        return False
    else:
        return True


def isint(x):
    try:
        a = float(x)
        b = int(a)
    except ValueError:
        return False
    else:
        return a == b


def islist(x):
    return True if x.count(',') > 0 else False


def isbool(x):
    return x.lower() in ("true", "false")


def str2type(x):
    try:
        if isint(x):
            return int(x)
        elif isfloat(x):
            return float(x)
        elif islist(x):
            return str2list(x)
        else:
            return x
    except:
        return x


def str2list(x):
    # remove list brackets from string
    if x.startswith('[') and x.endswith(']'):
        x = x[1:-1]

    try:
        x = [int(k) if isint(k) else float(k) if isfloat(k) else k for k in x.replace(' ', '').split(',')]
    except:
        return False

    return x


def rint(val):
    return int(round(val))



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

from plenopticam.lfp_extractor import LfpViewpoints
from plenopticam.misc import img_resize, save_img_file

from depthy.lightfield import epi_depth
from depthy.misc import disp2pts, save_ply, save_pfm, plot_point_cloud
from os.path import join
import warnings
from typing import Union

try:
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.pyplot as plt
except ImportError as e:
    warnings.warn('matplotlib is not installed')
    Axes3D = None


class LfpDepth(LfpViewpoints):

    def __init__(self, *args, **kwargs):
        super(LfpDepth, self).__init__(*args, **kwargs)

        # initialization
        self.depth_map = kwargs['depth_map'] if 'depth_map' in kwargs else None

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # print status
        self.sta.status_msg('Compute depth map', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        # compute depth
        self.depth_map = epi_depth(lf_img_arr=self.vp_img_arr.copy(), lf_wid=1, primal_opt=True, perc_clip=1)
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

        # export data
        self.sta.status_msg('Write depth data', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])
        pts = disp2pts(self.depth_map, rgb_img=self.central_view.copy(), focus_dist_mm=200)
        save_ply(pts, file_path=join(self.cfg.exp_path, 'depth.ply'))
        save_pfm(self.depth_map, scale=1, file_path=join(self.cfg.exp_path, 'depth.pfm'))
        save_img_file(self.depth_map, file_path=join(self.cfg.exp_path, 'depth.png'))
        self.sta.progress(100, self.cfg.params[self.cfg.opt_prnt])

    def plot_point_cloud(self,
                         rgb_img=None,
                         down_scale: int = 4,
                         view_angles: Union[int, int] = (50, 70),
                         s: float = 0.5,
                         ax: Axes3D = None) -> Axes3D:

        # use valid central view if rgb image is missing
        rgb_img = self.central_view if rgb_img is None and self.central_view is not None else rgb_img

        # downsample via interpolation
        if down_scale >= 1:
            rgb_img = img_resize(rgb_img.copy(), x_scale=1./down_scale) if rgb_img is not None else None
            dpt_map = img_resize(self.depth_map.copy(), x_scale=1./down_scale)
        else:
            warnings.warn('Downscale parameter has to be a positive integer greater than zero.')
            dpt_map = self.depth_map.copy()

        if self.depth_map is not None:
            ax = plot_point_cloud(dpt_map, rgb_img=rgb_img, down_scale=1, view_angles=view_angles, s=s, ax=ax)
        else:
            self.sta.status_msg(msg='Depth map variable is empty')
            ax = Axes3D(fig=plt.figure())

        return ax



# local imports
from plenopticam.lfp_aligner.lfp_microlenses import LfpMicroLenses
from plenopticam.lfp_calibrator.grid_fitter import GridFitter
from plenopticam import misc

# external libs
import numpy as np
from skimage import transform
import copy


class LfpGlobalResampler(LfpMicroLenses):

    def __init__(self, *args, **kwargs):
        super(LfpGlobalResampler, self).__init__(*args, **kwargs)

        # internal variables
        self._lfp_prj = self._lfp_img.copy()

    def global_resampling(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # global light-field image rectification
        self.projective_alignment()

        # global rectification of hexagonal micro image array
        self.hexagonal_alignment()

    def projective_alignment(self):

        # get projective matrix from centroids
        gfxy = GridFitter(self._CENTROIDS.copy(), flip_xy=True)
        gfxy.main()
        pmat_cent = gfxy.pmat.copy()
        R, K, t = gfxy.decompose(pmat_cent)
        del gfxy

        # get target micro image size
        target_pitch = np.array([np.floor(K[1, 1]), np.floor(K[0, 0])])
        if self.cfg.calibs[self.cfg.pat_type] == 'rec':
            # odd micro image size if rectangular
            target_pitch += np.mod(target_pitch+1, 2)
        elif self.cfg.calibs[self.cfg.pat_type] == 'hex':
            # even micro image size if hexagonal
            target_pitch += np.mod(target_pitch, 2)

        self._limg_pitch = target_pitch.astype('int')
        scales = np.array([self._limg_pitch[0]/K[1, 1], self._limg_pitch[1]/K[0, 0]])
        oshape = np.round(np.array(self._lfp_img.shape[:2])*scales).astype('int')

        # form scale and translation matrix
        Kpts = np.diag([self._limg_pitch[0], self._limg_pitch[1], K[2, 2]])

        # stretch axis with higher sampling density in hexagonal case for
        Kpts[0, 0] = Kpts[0, 0]*2/np.sqrt(3) if self.cfg.calibs[self.cfg.pat_type] == 'hex' else Kpts[0, 0]

        # create projective matrix of desired grid
        dims = [int(max(self._CENTROIDS[:, 2])+1), int(max(self._CENTROIDS[:, 3])+1)]
        grid = GridFitter.grid_gen(dims, pat_type=self.cfg.calibs[self.cfg.pat_type], hex_odd=self._hex_odd)
        grid = GridFitter.apply_transform(Kpts, grid)

        # shift grid
        shifts = np.zeros(2)
        cropsets = self._limg_pitch // 2
        shifts[0] = min(grid[:, 0]) - cropsets[0] if min(grid[:, 0]) - cropsets[0] < 0 else 0
        shifts[1] = min(grid[:, 1]) - cropsets[1] if min(grid[:, 1]) - cropsets[1] < 0 else 0
        grid[:, :2] = grid[:, :2] - shifts

        grxy = GridFitter(grid.copy(), flip_xy=True)
        grxy.main()
        pmat_grid = grxy.pmat.copy()
        del grxy

        # compute transfer matrix
        tra_mat = np.dot(pmat_grid, np.linalg.inv(pmat_cent))

        # transform light-field image
        tform = transform.ProjectiveTransform(matrix=tra_mat)
        self._lfp_prj = transform.warp(self._lfp_img, tform.inverse, output_shape=oshape)

    def hexagonal_alignment(self):

        if self._lfp_prj is None or not isinstance(self._lfp_prj, np.ndarray):
            return False

        if 0 in self._limg_pitch:
            raise Exception('Micro image size cannot be zero.')

        if self.cfg.calibs[self.cfg.pat_type] == 'rec':
            self._lfp_img_align = self._lfp_prj
            return True

        dims = self._lfp_prj.shape
        lens_new_x = int(round(self._LENS_X_MAX*2/np.sqrt(3)))
        pad_zro = np.zeros([*self._limg_pitch, dims[2]])
        pad_hlf = np.zeros([self._limg_pitch[0], self._limg_pitch[1]//2, dims[2]])
        hex_switch = copy.copy(self._hex_odd)
        lnew_pitch = self._limg_pitch + np.mod(self._limg_pitch+1, 2)

        # add one in micro image size for odd
        self._lfp_img_align = np.zeros([int((self._LENS_Y_MAX-1)*(self._limg_pitch[0]+1)),
                                        int(lens_new_x*(self._limg_pitch[1]+1)),
                                        dims[2]])

        for ly in range(self._LENS_Y_MAX-1):

            # print progress status for on console
            self.sta.progress((ly+1)/self._LENS_Y_MAX*100, self.cfg.params[self.cfg.opt_prnt])

            if hex_switch == 1:
                row_a = np.hstack(
                    [pad_hlf,
                     self._lfp_prj[ly*self._limg_pitch[0]:(ly+1)*self._limg_pitch[0], :-self._limg_pitch[1]//2, :]])
                row_b = self._lfp_prj[(ly+1)*self._limg_pitch[0]:(ly+2)*self._limg_pitch[0], ...]
                row_c = np.hstack([row_a[:, self._limg_pitch[1]:, :], pad_zro])
            else:
                row_a = self._lfp_prj[ly*self._limg_pitch[0]:(ly+1)*self._limg_pitch[0], ...]
                row_b = np.hstack(
                    [pad_hlf,
                     self._lfp_prj[(ly+1)*self._limg_pitch[0]:(ly+2)*self._limg_pitch[0], :-self._limg_pitch[1]//2, :]])
                row_c = np.hstack([row_b[:, self._limg_pitch[1]:, :], pad_zro])

            row_tensor = np.concatenate([row_a[..., None], row_b[..., None], row_c[..., None]], axis=-1)
            row_interp = np.mean(row_tensor, axis=-1)

            # elongation to compensate for hexagonal aspect ratio
            row_string = self.hex_stretch(row_interp)

            # resize row for odd micro image pitch
            row_string = misc.img_resize(row_string,
                                         x_scale=lnew_pitch[1] / self._limg_pitch[1],
                                         y_scale=lnew_pitch[0] / self._limg_pitch[0])

            # place micro image row into aligned image
            self._lfp_img_align[ly*(self._limg_pitch[0]+1):(ly+1)*(self._limg_pitch[0]+1), ...] = row_string

            # flip hexagonal shift orientation every other micro lens row
            hex_switch = not hex_switch

        # re-evaluate new odd micro image size
        self._limg_pitch = lnew_pitch

        # skip first and last column of half micro images
        self._lfp_img_align = self._lfp_img_align[:, self._limg_pitch[1]//2+1:-self._limg_pitch[1]//2, :]

        return self._lfp_img_align

    def hex_stretch(self, lf_row):

        dims = lf_row.shape
        lens_new_x = int(round(self._LENS_X_MAX * 2 / np.sqrt(3)))
        interp_stack = np.zeros([dims[0], lens_new_x * self._limg_pitch[1], dims[2]])

        # image stretch interpolation in x-direction to compensate for hex-alignment
        for y in range(self._limg_pitch[0]):
            for x in range(self._limg_pitch[1]):
                lf_row_pos = lf_row[y, x::self._limg_pitch[1], :]
                for p in range(dims[2]):
                    # stack of micro images elongated in x-direction
                    interp_coords = np.linspace(0, self._LENS_X_MAX, lens_new_x)
                    interp_string = np.interp(interp_coords, range(self._LENS_X_MAX), lf_row_pos[:self._LENS_X_MAX, p])
                    interp_stack[y, x::self._limg_pitch[1], p] = interp_string

        return interp_stack



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from plenopticam.gui.constants import PX, PY, BTN_W


# make object for command widget
class CmndWidget(tk.Frame):

    def __init__(self, parent):

        # inheritance
        tk.Frame.__init__(self, parent)
        self.parent = parent
        self.cfg = parent.cfg
        self.btn_list = []

        # button for light field processing
        run_btn = tk.Button(self, text='Process', width=BTN_W, command=self.parent.process)
        run_btn.grid(row=0, column=0, padx=PX, pady=PY)
        self.btn_list.append(run_btn)

        # button for settings configuration
        cfg_btn = tk.Button(self, text='Settings', width=BTN_W, command=self.parent.cfg_change)
        cfg_btn.grid(row=0, column=1, padx=PX, pady=PY)
        self.btn_list.append(cfg_btn)

        # button to stop/cancel process
        viw_btn = tk.Button(self, text='Viewer', width=BTN_W, command=self.parent.view)
        viw_btn.grid(row=0, column=2, padx=PX, pady=PY)
        self.btn_list.append(viw_btn)

        # button for application shutdown
        qit_btn = tk.Button(self, text='Quit', width=BTN_W, command=self.parent.exit)
        qit_btn.grid(row=0, column=3, padx=PX, pady=PY)
        self.btn_list.append(qit_btn)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""


# local imports
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import safe_get, PlenopticamStatus
from plenopticam.lfp_reader.constants import SUPP_FILE_EXT

# external libs
import json
import hashlib
import numpy as np
import os


class LfpDecoder(object):

    # static class variables
    LFP_HEADER = b'\x89\x4c\x46\x50\x0d\x0a\x1a\x0a\x00\x00\x00\x01'  # LFP header
    LFM_HEADER = b'\x89\x4c\x46\x4d\x0d\x0a\x1a\x0a\x00\x00\x00\x00'  # table of contents header
    LFC_HEADER = b'\x89\x4c\x46\x43\x0d\x0a\x1a\x0a\x00\x00\x00\x00'  # content section header
    PADDING_LEN = 4
    SHA1_LEN = 45
    SHA_PADDING_LEN = 35

    def __init__(self, file=None, cfg=None, sta=None, **kwargs):

        # input variables
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()
        self.file = file
        self._lfp_path = kwargs['lfp_path'] if 'lfp_path' in kwargs else self.cfg.params[self.cfg.lfp_path]

        # internal variables
        self._json_dict = kwargs['json_dict'] if 'json_dict' in kwargs else {}
        self._shape = None
        self._img_buf = None

        # output variable
        self.cfg.lfpimg = {} if not hasattr(self.cfg, 'lfpimg') else self.cfg.lfpimg
        self._bay_img = None

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # LFC type decoding
        if self._lfp_path.lower().endswith(SUPP_FILE_EXT[:2]):
            self.decode_lfc()

        # C.x bundle type decoding
        elif self._lfp_path.lower().endswith(SUPP_FILE_EXT[2:]):
            self.decode_bundle()

        # raw type decoding
        elif self._lfp_path.lower().endswith(SUPP_FILE_EXT[2]):
            self.decode_raw()

    def decode_lfc(self):

        # decode lfp file
        sections = self.read_buffer(self.file)

        # retrieve JSON data
        self._json_dict = self.read_json(sections)

        # JSON file export
        dp = os.path.splitext(self._lfp_path)[0]
        self.cfg.save_json(os.path.join(dp, os.path.basename(dp) + '.json'), json_dict=self.json_dict)

        # validate camera format support
        if not self.valid_cam_type:
            return False

        # decompose JSON data
        self._shape = [safe_get(self._json_dict, 'image', 'width'), safe_get(self._json_dict, 'image', 'height')]

        # filter LFP metadata settings
        self.cfg.lfpimg = self.filter_lfp_json(self._json_dict)

        # compose bayer image from lfp file
        sec_idx = self.get_idx(sections, int(self._shape[0] * self._shape[1] * self.cfg.lfpimg['bit'] / 8))[0]
        self._img_buf = list(sections[sec_idx])
        self.comp_bayer()

        return True

    def decode_raw(self):

        # read bytes from file
        self._img_buf = list(self.file) if isinstance(self.file, bytes) else list(self.file.read())

        if len(self._img_buf) >= int(7728*5368*10/8):
            self.cfg.lfpimg['bit'] = 10
            self._shape = [7728, 5368]
            self.cfg.lfpimg['bay'] = 'GRBG'
        elif len(self._img_buf) >= int(3280**2*10/8):
            self.cfg.lfpimg['bit'] = 12
            self._shape = [3280, 3280]
            self.cfg.lfpimg['bay'] = 'BGGR'
        else:
            self.sta.status_msg('File type not recognized')
            self.sta.error = True
            return False

        # compose bayer image from lfp file
        self.comp_bayer()

        return True

    def decode_bundle(self):

        # decode lfp file
        sections = self.read_buffer(self.file)

        # retrieve JSON data
        self._json_dict = self.read_json(sections)

        # JSON file export
        dp = os.path.splitext(self._lfp_path)[0]
        self.cfg.save_json(os.path.join(dp, os.path.basename(self._lfp_path) + '.json'), json_dict=self.json_dict)

        # decompose packed files from calibration bundle
        FILE_NUM = len(self._json_dict['files'])
        for idx, (section, file_dict) in enumerate(zip(sections[-FILE_NUM:], self._json_dict['files'])):
            # extract file name and write file
            fn = file_dict['name'].split('\\')[-1]
            with open(os.path.join(dp, fn), 'w') as f:
                try:
                    f.write(section.decode('utf-8'))
                except:
                    f.write(str(section))

        return True

    @property
    def valid_cam_type(self):
        """ check if Lytro file format is supported """

        # search for 2nd generation keys (filter camera serial and model )
        serial = safe_get(self._json_dict, "camera", "serialNumber")
        cam_model = serial if serial else safe_get(self._json_dict, "camera", "model")

        # search for 1st generations keys (file names)
        #files = safe_get(self._json_dict, "files")

        if cam_model is None:# and files is None:
            self.sta.status_msg('File type not recognized')
            self.sta.error = True
            return False

        return True

    @staticmethod
    def filter_lfp_json(json_dict, settings=None):
        """ filter LFP metadata settings """

        # variable init
        settings = {} if settings is None else settings
        channels = ['b', 'r', 'gb', 'gr']

        # filter camera serial and model
        serial = safe_get(json_dict, 'camera', 'serialNumber')
        cam_model = serial if serial else safe_get(json_dict, 'camera', 'model')

        # set decode paramaters considering camera model
        if cam_model.startswith(('A', 'F')):    # 1st generation Lytro

            # read bit packing
            settings['bit'] = safe_get(json_dict, "image", "rawDetails", "pixelPacking", "bitsPerPixel")
            if not settings['bit'] == 12:
                raise AssertionError('Unrecognized bit packing format')

            # get Bayer pattern, Automatic White Balance (AWB) gains and Color Correction Matrix (CCM)
            settings['bay'] = 'BGGR'
            settings['awb'] = [safe_get(json_dict, 'image', 'color', 'whiteBalanceGain', key) for key in channels]
            settings['ccm'] = safe_get(json_dict, 'image', 'color', 'ccmRgbToSrgbArray')
            settings['gam'] = safe_get(json_dict, 'image', 'color', 'gamma')

        elif cam_model.startswith(('B', 'I')) or cam_model.isdigit():  # 2nd generation Lytro

            # read bit packing
            settings['bit'] = safe_get(json_dict, "image", "pixelPacking", "bitsPerPixel")
            if not settings['bit'] == 10:
                raise AssertionError('Unrecognized bit packing format')

            # get Bayer pattern, Automatic White Balance (AWB) gains and Color Correction Matrix (CCM)
            settings['bay'] = 'GRBG'
            settings['awb'] = [safe_get(json_dict, 'algorithms', 'awb', 'computed', 'gain', key) for key in channels]
            settings['ccm'] = safe_get(json_dict, 'image', 'color', 'ccm')
            settings['gam'] = safe_get(json_dict, 'master', 'picture', 'frameArray', 0, 'frame', 'metadata', 'image',
                                       'color', 'gamma')
            settings['exp'] = safe_get(json_dict, "image", "modulationExposureBias")

        return settings

    def read_buffer(self, f):

        # list init
        sections = []

        try:
            # header decomposition and validation checks
            file_header = f.read(len(self.LFP_HEADER))
            if not file_header == self.LFP_HEADER:
                raise AssertionError('File header type not recognized')
            header_length = int(f.read(4).hex(), 16)
            if not header_length == 0:
                raise AssertionError('Unexpected header length')
            while f.read(1) != b'':
                f.seek(-1, 1)   # move back one byte
                sections.append(self.read_section(f))
            f.close()
        except AssertionError:
            f.close()

        return sections

    def read_section(self, f):

        # read section header
        sect_header = f.read(len(self.LFM_HEADER))
        if not (sect_header == self.LFM_HEADER or sect_header == self.LFC_HEADER):
            raise AssertionError('Section header type not recognized')

        # read data section length
        sect_len = int(f.read(4).hex(), 16)

        # read sha1 checksum and padding
        sha1 = f.read(self.SHA1_LEN)
        padding = int(f.read(self.SHA_PADDING_LEN).hex(), 16)   # skip padding
        if not padding == 0:
            raise AssertionError('Unexpected padding length')

        # read data section and evaluate sha-1 checksum
        section = f.read(sect_len)
        if not sha1[5:] == hashlib.sha1(section).hexdigest().encode('utf-8'):
            raise AssertionError('Corrupted section %s' % section)

        # move forward to next section while ignoring padded bytes
        while True:
            padding = f.read(1)
            if padding == '' or padding != b'\x00':
                break

        if padding != b'':
            f.seek(-1, 1)   # move back one byte

        return section

    def comp_bayer(self):
        """ inspired by Nirav Patel's lfptools """

        # initialize column vector for bayer image array
        self._bay_img = np.zeros(self._shape[0]*self._shape[1], dtype=np.uint16)

        # determine bit packing
        bit_pac = self.cfg.lfpimg['bit'] if 'bit' in self.cfg.lfpimg.keys() else 10

        if bit_pac == 10:

            t0 = np.array(self._img_buf[0::5], 'uint16')
            t1 = np.array(self._img_buf[1::5], 'uint16')
            t2 = np.array(self._img_buf[2::5], 'uint16')
            t3 = np.array(self._img_buf[3::5], 'uint16')
            t4 = np.array(self._img_buf[4::5], 'uint16')

            t0 = t0 << 2
            t1 = t1 << 2
            t2 = t2 << 2
            t3 = t3 << 2

            t0 += (t4 & 0x03)
            t1 += (t4 & 0x0C) >> 2
            t2 += (t4 & 0x30) >> 4
            t3 += (t4 & 0xC0) >> 6

            self._bay_img = np.empty((4*t0.size,), dtype=t0.dtype)
            self._bay_img[0::4] = t0        # green-red
            self._bay_img[1::4] = t1        # red
            self._bay_img[2::4] = t2        # green-blue
            self._bay_img[3::4] = t3        # blue

        elif bit_pac == 12:

            t0 = np.array(self._img_buf[0::3], 'uint16')
            t1 = np.array(self._img_buf[1::3], 'uint16')
            t2 = np.array(self._img_buf[2::3], 'uint16')

            a0 = (t0 << 4) + ((t1 & 0xF0) >> 4)
            a1 = ((t1 & 0x0F) << 8) + t2

            self._bay_img = np.empty((2*a0.size,), dtype=t0.dtype)
            self._bay_img[0::4] = a0[0::2]  # blue
            self._bay_img[2::4] = a0[1::2]  # red
            self._bay_img[1::2] = a1        # green

        # rearrange column vector to 2-D image array
        self._bay_img = np.reshape(self._bay_img, (self._shape[1], self._shape[0]))

        # convert to float
        self._bay_img = self._bay_img.astype('float')

        return True

    @staticmethod
    def read_json(sections):
        json_dict = {}
        for i in range(len(sections)):
            try:
                json_dict.update(json.loads(sections[i].decode('utf-8')))
            except UnicodeDecodeError:
                pass
            except json.decoder.JSONDecodeError:
                pass

        return json_dict

    @staticmethod
    def get_idx(checklist, value):
        return [x for x in range(len(checklist)) if len(checklist[x]) == value]

    @property
    def bay_img(self):
        return (self._bay_img.copy()-65)/(1023-65) if self._bay_img is not None else None

    @property
    def json_dict(self):
        return self._json_dict.copy()



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

# local imports
from plenopticam.cfg import PlenopticamConfig
from plenopticam import misc
from plenopticam.lfp_extractor.lfp_cropper import LfpCropper
from plenopticam.lfp_extractor.lfp_rearranger import LfpRearranger
from plenopticam.lfp_extractor.lfp_exporter import LfpExporter
from plenopticam.lfp_extractor.lfp_contrast import LfpContrast
from plenopticam.lfp_extractor.lfp_outliers import LfpOutliers
from plenopticam.lfp_extractor.lfp_color_eq import LfpColorEqualizer
from plenopticam.lfp_extractor.hex_corrector import HexCorrector
from plenopticam.lfp_extractor.lfp_depth import LfpDepth

import pickle
import os


class LfpExtractor(object):

    def __init__(self, lfp_img_align=None, cfg=None, sta=None):

        # input variables
        self._lfp_img_align = lfp_img_align
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else misc.PlenopticamStatus()

        # variables for viewpoint arrays
        self.vp_img_arr = []        # gamma corrected
        self.vp_img_linear = []     # linear gamma (for further processing)
        self.depth_map = None

    def main(self):

        # load previously calculated calibration and aligned data
        if self.cfg.calibs is None:
            self.cfg.load_cal_data()
        if self._lfp_img_align is None:
            self.load_pickle_file()
            self.load_lfp_metadata()

        # micro image crop
        lfp_obj = LfpCropper(lfp_img_align=self._lfp_img_align, cfg=self.cfg, sta=self.sta)
        lfp_obj.main()
        self._lfp_img_align = lfp_obj.lfp_img_align
        del lfp_obj

        # rearrange light-field to sub-aperture images
        if self.cfg.params[self.cfg.opt_view]:
            lfp_obj = LfpRearranger(self._lfp_img_align, cfg=self.cfg, sta=self.sta)
            lfp_obj.main()
            self.vp_img_linear = lfp_obj.vp_img_arr
            del lfp_obj

        # remove outliers if option is set
        if self.cfg.params[self.cfg.opt_lier]:
            obj = LfpOutliers(vp_img_arr=self.vp_img_linear, cfg=self.cfg, sta=self.sta)
            obj.main()
            self.vp_img_linear = obj.vp_img_arr
            del obj

        # color equalization
        if self.cfg.params[self.cfg.opt_colo]:
            obj = LfpColorEqualizer(vp_img_arr=self.vp_img_linear, cfg=self.cfg, sta=self.sta)
            obj.main()
            self.vp_img_linear = obj.vp_img_arr
            del obj

        # copy light-field before gamma encoding for refocusing process (prior to contrast and export)
        self.vp_img_arr = self.vp_img_linear.copy() if self.vp_img_linear is not None else None

        # color management automation
        obj = LfpContrast(vp_img_arr=self.vp_img_arr, cfg=self.cfg, sta=self.sta)
        obj.main()
        self.vp_img_arr = obj.vp_img_arr
        del obj

        # reduction of hexagonal sampling artifacts
        if self.cfg.params[self.cfg.opt_arti]:
            obj = HexCorrector(vp_img_arr=self.vp_img_arr, cfg=self.cfg, sta=self.sta)
            obj.main()
            self.vp_img_arr = obj.vp_img_arr
            del obj

        # write viewpoint data to hard drive
        if self.cfg.params[self.cfg.opt_view]:
            obj = LfpExporter(vp_img_arr=self.vp_img_arr, cfg=self.cfg, sta=self.sta)
            obj.write_viewpoint_data()
            del obj

        # compute and write depth data from epipolar analysis
        if self.cfg.params[self.cfg.opt_dpth]:
            obj = LfpDepth(vp_img_arr=self.vp_img_arr, cfg=self.cfg, sta=self.sta)
            obj.main()
            self.depth_map = obj.depth_map
            del obj

        return True

    def load_pickle_file(self):
        """ load previously computed light field alignment """

        # file path
        fp = os.path.join(self.cfg.exp_path, 'lfp_img_align.pkl')

        try:
            self._lfp_img_align = pickle.load(open(fp, 'rb'))
        except EOFError:
            os.remove(fp)
        except FileNotFoundError:
            return False

        return True

    def load_lfp_metadata(self):
        """ load LFP metadata settings (for Lytro files only) """

        fname = os.path.splitext(os.path.basename(self.cfg.params[self.cfg.lfp_path]))[0]+'.json'
        fp = os.path.join(self.cfg.exp_path, fname)
        if os.path.isfile(fp):
            json_dict = self.cfg.load_json(fp=fp, sta=None)
            from plenopticam.lfp_reader.lfp_decoder import LfpDecoder
            self.cfg.lfpimg = LfpDecoder().filter_lfp_json(json_dict, settings=self.cfg.lfpimg)

        return True



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import unittest

import pickle
from os.path import join, exists, basename
from os import listdir

from plenopticam.lfp_reader import LfpReader
from plenopticam.lfp_calibrator import LfpCalibrator, CaliFinder, CentroidDrawer
from plenopticam.lfp_aligner import LfpAligner
from plenopticam.lfp_extractor import LfpExtractor
from plenopticam.lfp_refocuser import LfpRefocuser
from plenopticam.cfg import PlenopticamConfig, constants
from plenopticam.misc import PlenopticamStatus, mkdir_p
from plenopticam.misc.data_downloader import DataDownloader


class PlenoptiCamTesterIllum(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(PlenoptiCamTesterIllum, self).__init__(*args, **kwargs)

    def setUp(self):

        # instantiate config and status objects
        self.cfg = PlenopticamConfig()
        self.cfg.default_values()
        self.sta = PlenopticamStatus()

        # enable options in config to cover more algorithms in tests
        self.cfg.params[self.cfg.cal_meth] = constants.CALI_METH[3]
        self.cfg.params[self.cfg.opt_cali] = True
        self.cfg.params[self.cfg.opt_vign] = True
        self.cfg.params[self.cfg.opt_rota] = True
        self.cfg.params[self.cfg.opt_refi] = True
        self.cfg.params[self.cfg.opt_pflu] = True
        self.cfg.params[self.cfg.opt_arti] = True
        self.cfg.params[self.cfg.opt_lier] = True
        self.cfg.params[self.cfg.opt_cont] = True
        self.cfg.params[self.cfg.opt_awb_] = True
        self.cfg.params[self.cfg.opt_sat_] = True
        self.cfg.params[self.cfg.opt_dbug] = True
        self.cfg.params[self.cfg.opt_dpth] = True
        self.cfg.params[self.cfg.ran_refo] = [0, 1]

        # compute 3x3 viewpoints only (to reduce computation time)
        self.cfg.params[self.cfg.ptc_leng] = 7

        # skip progress prints (prevent Travis from terminating due to reaching 4MB logfile size)
        self.sta.prog_opt = False

        # print current process message (to prevent Travis from stopping after 10 mins)
        self.cfg.params[self.cfg.opt_prnt] = True

        # retrieve Lytro Illum data
        self.loader = DataDownloader(cfg=self.cfg, sta=self.sta)
        self.fp = join(self.loader.root_path, 'examples', 'data')
        archive_fn = join(self.fp, basename(self.loader.host_eu_url))
        self.loader.download_data(self.loader.host_eu_url, fp=self.fp) if not exists(archive_fn) else None
        self.loader.extract_archive(archive_fn, fname_list='lfr')

    def test_illum(self):

        # use pre-loaded calibration dataset
        wht_list = [file for file in listdir(self.fp) if file.startswith('caldata')]
        lfp_list = [file for file in listdir(self.fp) if file.endswith(('lfr', 'lfp'))]

        self.cfg.params[self.cfg.cal_path] = join(self.fp, wht_list[0])

        for lfp_file in lfp_list:

            self.cfg.params[self.cfg.lfp_path] = join(self.fp, lfp_file)
            print('\nCompute image %s' % basename(self.cfg.params[self.cfg.lfp_path]))

            # decode light field image
            obj = LfpReader(self.cfg, self.sta)
            ret = obj.main()

            # use third of original image size (to prevent Travis from stopping due to memory error)
            crop_h, crop_w = obj.lfp_img.shape[0] // 3, obj.lfp_img.shape[1] // 3
            crop_h, crop_w = crop_h + crop_h % 2, crop_w + crop_w % 2   # use even number for correct Bayer arrangement
            lfp_img = obj.lfp_img[crop_h:-crop_h, crop_w:-crop_w]
            del obj

            self.assertEqual(True, ret)

            # create output data folder
            mkdir_p(self.cfg.exp_path, self.cfg.params[self.cfg.opt_prnt])

            if not self.cfg.cond_meta_file():
                # automatic calibration data selection
                obj = CaliFinder(self.cfg, self.sta)
                ret = obj.main()
                wht_img = obj.wht_bay[crop_h:-crop_h, crop_w:-crop_w] if obj.wht_bay is not None else obj.wht_bay
                del obj

                self.assertEqual(True, ret)

            meta_cond = not (exists(self.cfg.params[self.cfg.cal_meta]) and self.cfg.params[self.cfg.cal_meta].lower().endswith('json'))
            if meta_cond or self.cfg.params[self.cfg.opt_cali]:
                # perform centroid calibration
                obj = LfpCalibrator(wht_img, self.cfg, self.sta)
                ret = obj.main()
                self.cfg = obj.cfg
                del obj

                self.assertEqual(True, ret)

            # load calibration data
            self.cfg.load_cal_data()

            # write centroids as png file
            if wht_img is not None:
                obj = CentroidDrawer(wht_img, self.cfg.calibs[self.cfg.mic_list], self.cfg)
                ret = obj.write_centroids_img(fn='testcase_wht_img+mics.png')
                del obj

                self.assertEqual(True, ret)

            #  check if light field alignment has been done before
            if self.cfg.cond_lfp_align():
                # align light field
                obj = LfpAligner(lfp_img, self.cfg, self.sta, wht_img)
                ret = obj.main()
                del obj

                self.assertEqual(True, ret)

            # load previously computed light field alignment
            with open(join(self.cfg.exp_path, 'lfp_img_align.pkl'), 'rb') as f:
                lfp_img_align = pickle.load(f)

            # extract viewpoint data
            CaliFinder(self.cfg).main()
            obj = LfpExtractor(lfp_img_align, cfg=self.cfg, sta=self.sta)
            ret = obj.main()
            vp_img_arr = obj.vp_img_linear
            del obj

            self.assertEqual(True, ret)

            # do refocusing
            if self.cfg.params[self.cfg.opt_refo]:
                obj = LfpRefocuser(vp_img_arr, cfg=self.cfg, sta=self.sta)
                ret = obj.main()
                del obj

            self.assertEqual(True, ret)

        return True

    def test_all(self):

        self.test_illum()


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import division

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2017 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

# local imports
from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc.status import PlenopticamStatus
from plenopticam.lfp_calibrator.centroid_drawer import CentroidDrawer

DR = 1


class CentroidRefiner(object):

    def __init__(self, img, centroids, cfg=None, sta=None, M=None, method=None):
        super(CentroidRefiner, self).__init__()
        """

        This class takes a list of integer coordinates as centroids and an intensity map as inputs and re-computes

        :param img: image used as basis for coordinate refinement calculation
        :param centroids: iterable (list or np.ndarray) with coordinates of integer type
        :param cfg: PlenoptiCam configuration object
        :param sta: PlenoptiCam status object
        :param M: micro image size
        :param method: 'area' or 'peak'
        """

        # input variables
        self._img = np.asarray(img)
        self._centroids = np.array(centroids)
        self.cfg = cfg if cfg is not None else PlenopticamConfig()
        self.sta = sta if sta is not None else PlenopticamStatus()
        self._M = M
        self._r = self._M//2//DR if M is not None else None
        self._method = cfg.params[cfg.cal_meth] if cfg is not None else method if method is not None else 'area'

        # internal variables
        self._t = self._s = 0

        # output variables
        self._centroids_refined = []

    def main(self):

        # check interrupt status
        if self.sta.interrupt:
            return False

        # status message handling
        self.sta.status_msg('Refine micro image centers', self.cfg.params[self.cfg.opt_prnt])
        self.sta.progress(None, self.cfg.params[self.cfg.opt_prnt])

        fun = self._peak_centroid if self._method == 'peak' else self._area_centroid

        # image normalization
        self._img = (self._img - self._img.min()) / (self._img.max() - self._img.min())

        # coordinate and image downsampling preparation
        self._centroids = [(x//DR, y//DR) for x, y in self._centroids] if DR > 1 else self._centroids
        img_scale = self._img[::DR, ::DR]

        self._centroids_refined = []

        # iterate through all centroids for refinement procedure
        for i, m in enumerate(self._centroids):

            # check interrupt status
            if self.sta.interrupt:
                return False

            # compute refined coordinates based on given method
            fun(img_scale[int(m[0])-self._r:int(m[0])+self._r+1, int(m[1])-self._r:int(m[1])+self._r+1], m)
            self._centroids_refined.append(self._get_coords())

            # print status
            self.sta.progress((i+1)/len(self._centroids)*100, self.cfg.params[self.cfg.opt_prnt])

        # coordinate upsampling to compensate for downsampling
        self._centroids_refined = [(x*DR, y*DR) for x, y in self._centroids_refined] if DR > 1 else self._centroids_refined

        # remove centroids at image boundaries
        self.exclude_marginal_centroids()

        # write centroids image to hard drive (if option set)
        if self.cfg.params[self.cfg.opt_dbug] and not self.sta.interrupt:
            draw_obj = CentroidDrawer(self._img, self._centroids, self.cfg)
            draw_obj.write_centroids_img(fn='wht_img+mics_refi.png')
            del draw_obj

        return True

    def _thresholding(self, input_win):

        # parameter init
        weight_win = input_win / input_win.max()

        # window thresholding
        th_img = np.zeros_like(input_win, dtype='bool')
        th_val = np.percentile(weight_win, 75)
        th_img[weight_win > th_val] = 1

        # validate that binary region exists
        if np.sum(th_img) == 0:
            raise Exception('Binary object not found')

        return th_img

    def _area_centroid(self, input_win, p):

        th_img = self._thresholding(input_win)

        # binary centroid calculation in window
        count = (th_img == 1).sum()
        self._t, self._s = np.argwhere(th_img == 1).sum(0)/count if count > 0 else (float('nan'), float('nan'))
        self._t += p[0]-self._r
        self._s += p[1]-self._r

        return True

    def _peak_centroid(self, input_win, p, threshold_opt=False):

        # parameter init
        input_win *= self._thresholding(input_win) if threshold_opt is True else 1
        weight_win = input_win / input_win.sum()
        self._t = self._s = 0

        # generate coordinate indices
        iss = np.tile(np.arange(weight_win.shape[0]), weight_win.shape[1])
        jss = np.tile(np.arange(weight_win.shape[1]), weight_win.shape[0])

        # compute moment
        self._t = sum((iss+p[0]-self._r) * weight_win.T.flatten())
        self._s = sum((jss+p[1]-self._r) * weight_win.flatten())

        return True

    def exclude_marginal_centroids(self):
        """ remove centroids being closer to image border than half the micro image size M """

        h, w = self._img.shape[:2]

        if isinstance(np.version.version, str):
            # use numpy vectorization
            self._centroids_refined = np.array(self._centroids_refined)
            valid_idx = (self._r < self._centroids_refined[:, 0]) & (self._centroids_refined[:, 0] < h - self._r) & \
                        (self._r < self._centroids_refined[:, 1]) & (self._centroids_refined[:, 1] < w - self._r)
            self._centroids_refined = list(self._centroids_refined[valid_idx])
        else:
            # inline for loop
            self._centroids_refined = [c for c in self._centroids_refined
                                       if self._r < c[0] < h-self._r and self._r < c[1] < w-self._r]

        return True

    def _get_coords(self):
        return self._t, self._s

    @property
    def centroids_refined(self):
        return np.asarray(self._centroids_refined)



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "info@christopherhahne.de"
__license__ = """
    Copyright (c) 2019 Christopher Hahne <info@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import sys, errno
from os import makedirs, remove, chmod, listdir
from os.path import isdir, isfile, expanduser, join

from plenopticam.gui.constants import GEN_IMG_EXTS
from plenopticam.misc.file_rw import load_img_file
from plenopticam.misc.normalizer import Normalizer


def mkdir_p(path, print_opt=False):

    try:
        makedirs(path, mode=0o777)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and isdir(path):
            if print_opt:
                print('\n Potential data loss as directory already exists.')
        #else:
        #    raise OSError('\n Could not create directory.')

    return True


def rmdir_p(path, print_opt=False):

    try:
        import shutil
        shutil.rmtree(path, onerror=remove_readonly)
    except:
        if print_opt:
            print('\n Directory {0} could not be removed'.format(path))

    return True


def remove_readonly(func, path, _):
    """ clear the readonly bit and reattempt the removal """

    import stat
    chmod(path, stat.S_IWRITE)

    func(path)


def rm_file(path, print_opt=False):

    try:
        if isfile(path):
            remove(path)
    except OSError as e:
        if print_opt:
            print("\n Error: %s - %s." % (e.filename, e.strerror))


def select_file(init_dir=None, title=''):
    """ get file path from tkinter dialog """

    # consider initial directory if provided
    init_dir = expanduser('~/') if not init_dir else init_dir

    # import tkinter while considering Python version
    try:
        if sys.version_info > (3, 0):
            from tkinter import Tk
            from tkinter.filedialog import askopenfilename
        else:
            from Tkinter import Tk
            from tkFileDialog import askopenfilename
    except ImportError:
        raise ImportError('Please install tkinter package.')

    # open window using tkinter
    root = Tk()
    root.withdraw()
    root.update()
    file_path = askopenfilename(initialdir=[init_dir], title=title)
    root.update()

    return file_path if file_path else None


def get_img_list(img_dir, vp=1):
    """ obtain list of images from provided directory path """

    dir_list = listdir(img_dir)
    dir_list.sort()
    img_list = []
    for i in dir_list:
        img_path = join(img_dir, i)
        ext = img_path.split('.')[::-1][0].lower()
        if ext in GEN_IMG_EXTS:

            # load image
            img = load_img_file(img_path)

            # convert to uint8 if necessary
            img = Normalizer(img).uint8_norm() if str(img.dtype) != 'uint8' else img

            # append to image list
            img_list.append((i, img))

    # sort image list by indices in file names
    img_tuples = sorted(img_list, key=lambda k: idx_str_sort(k[0], 1 if vp else 0))
    _, img_list = zip(*img_tuples)

    if vp:
        vp_dim = int(np.sqrt(len(img_list)))
        img_list = np.reshape(img_list, newshape=(vp_dim, vp_dim) + img_list[0].shape, order='C')

    return img_list


def idx_str_sort(s, mode=0):
    """ criteria for sort in lambda function calls """

    if mode:
        # viewpoint mode
        return [int(s.split('.')[0].split('_')[0]),
                int(s.split('.')[0].split('_')[1])]
    else:
        # refocus mode
        return int(s.split('.')[0])



#!/usr/bin/env python

__author__ = "Christopher Hahne"
__email__ = "inbox@christopherhahne.de"
__license__ = """
    Copyright (c) 2020 Christopher Hahne <inbox@christopherhahne.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

import unittest
from os.path import join, dirname, abspath, exists
from os import makedirs, utime, listdir

try:
    import tkinter as tk
except ImportError:
    import Tkinter as tk

from plenopticam.cfg import PlenopticamConfig
from plenopticam.misc import PlenopticamStatus
from plenopticam.misc import rmdir_p, rm_file

from plenopticam.gui.widget_about import AbtWidget
from plenopticam.gui.widget_cmnd import CmndWidget
from plenopticam.gui.widget_cnfg import CnfgWidget
from plenopticam.gui.widget_ctrl import CtrlWidget
from plenopticam.gui.widget_file import FileWidget
from plenopticam.gui.widget_menu import MenuWidget
from plenopticam.gui.widget_path import PathWidget
from plenopticam.gui.widget_pbar import PbarWidget
from plenopticam.gui.widget_view import ViewWidget
from plenopticam.gui.top_level import PlenopticamApp
from plenopticam.gui.constants import PX, PY


class TKinterTestCase(unittest.TestCase):
    """
    Following methods are the same for each GUI test, so they are combined to a dedicated class.
    """
    def setUp(self):
        self.root = tk.Tk()
        self.cfg = PlenopticamConfig()
        self.sta = PlenopticamStatus()
        self.root.cfg = self.cfg
        self.root.sta = self.sta
        self.set_file_path()
        self.pump_events()

    def set_file_path(self):

        # folder and path handling
        self.fp = join(dirname(dirname(abspath(__file__))), 'examples', 'data')
        makedirs(self.fp) if not exists(self.fp) else None
        self.dummy_fn = 'test_dummy.lfp'
        self.dummy_path = join(self.fp, self.dummy_fn)

    def create_dummy_file(self):
        """ create a dummy file for tests """

        # create dummy file with wrong file format
        with open(self.cfg.params[self.cfg.lfp_path], 'a'):
            utime(self.cfg.params[self.cfg.lfp_path], None)

    def remove_dummy_file(self):
        """ remove dummy data after test """

        rm_file(self.cfg.params[self.cfg.lfp_path])
        rmdir_p(self.cfg.exp_path)

    def tearDown(self):
        if self.root:
            self.root.destroy()
            self.pump_events()

    def pump_events(self):
        while self.root.dooneevent(tk._tkinter.ALL_EVENTS | tk._tkinter.DONT_WAIT):
            pass


class PlenoptiCamTesterGui(TKinterTestCase):

    def test_plenopticam_app(self):

        wid = PlenopticamApp()
        self.pump_events()
        wid.destroy()

    def test_about_widget(self):

        wid = AbtWidget()
        self.pump_events()
        wid.abt_widget.destroy()

    def test_cmnd_widget(self):

        wid = CmndWidget(CtrlWidget(self.root))
        self.pump_events()
        wid.destroy()

    def test_cnfg_widget(self):

        wid = CnfgWidget(self.root)
        self.pump_events()

        wid.btn.focus_set()
        wid.btn.event_generate('<Return>')
        wid.destroy()

    def test_ctrl_widget(self):

        wid = CtrlWidget(self.root)

        # test path fetching
        wid.fil_wid.lfp_wid.ent.delete(0, "end")
        wid.fil_wid.cal_wid.ent.delete(0, "end")
        wid.fil_wid.lfp_wid.ent.insert(0, self.dummy_path)
        wid.fil_wid.cal_wid.ent.insert(0, self.dummy_path)
        self.cfg.params[self.cfg.lfp_path] = ''
        self.cfg.params[self.cfg.cal_path] = ''
        wid.fetch_paths()
        self.assertEqual(self.dummy_path, self.cfg.params[self.cfg.lfp_path])
        self.assertEqual(self.dummy_path, self.cfg.params[self.cfg.cal_path])

        self.pump_events()
        wid.destroy()

    def test_file_widget(self):

        wid = FileWidget(self.root)
        self.pump_events()
        wid.destroy()

    def test_menu_widget(self):

        wid = MenuWidget(CtrlWidget(self.root))
        self.pump_events()
        wid.open_docs()
        wid.open_about_dialog()
        wid.destroy()

    def test_path_widget(self):

        wid = PathWidget(self.root)

        # test getter and setter
        wid.path = self.dummy_path
        self.assertEqual(self.dummy_path, wid.path)

        self.pump_events()
        wid.destroy()

    def test_pbar_widget(self):

        self.root.sta = self.sta
        wid = PbarWidget(self.root)
        wid.pack(fill='both', expand=True, side='bottom', padx=PX, pady=PY)

        proc_name = 'Test'
        wid.update_stat_msg(proc_name)
        self.assertEqual(proc_name, wid.tk_stat.__getitem__('text'))

        exp_val = 50
        wid.update_prog_bar(exp_val)
        self.assertEqual(exp_val, wid.tk_prog_bar['value'])

        exp_val = 0
        wid.update_prog_msg(exp_val)
        self.assertEqual(str(exp_val)+' %', wid.s.configure("LabeledProgressbar", 'text').strip())

        exp_val = '1'
        wid.update_prog_msg(exp_val)
        self.assertEqual(exp_val, wid.s.configure("LabeledProgressbar", 'text').strip())

        wid.destroy()

    def test_view_widget(self):

        # dummy button with state key
        btn = {'state': 'normal'}

        # provide path from previously computed data
        lfp_list = [file for file in listdir(self.fp) if file.endswith(('lfr', 'lfp'))]
        self.cfg.params[self.cfg.lfp_path] = join(self.fp, lfp_list[0])

        wid = ViewWidget(self.root, cfg=self.cfg, sta=self.sta, btn=btn)

        # verify view button has been disabled after view widget instantiation
        self.assertEqual(btn['state'], 'disabled')

        self.pump_events()
        wid.btn_mode.focus_set()
        wid.btn_auto.event_generate('<Return>')
        wid.btn_auto.focus_set()
        wid.btn_auto.event_generate('<Return>')
        self.pump_events()

        wid.destroy()

        # verify view button has been enabled
        self.assertEqual(btn['state'], 'normal')

    def test_all(self):

        self.test_plenopticam_app()
        self.test_about_widget()
        self.test_cmnd_widget()
        self.test_cnfg_widget()
        self.test_ctrl_widget()
        self.test_file_widget()
        self.test_menu_widget()
        self.test_path_widget()
        self.test_pbar_widget()
        self.test_view_widget()
